""" This is the Proxy storage element client """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Resources.Storage.Utilities import checkArgumentFormat
from DIRAC.Resources.Storage.StorageBase import StorageBase
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Core.DISET.TransferClient import TransferClient
from DIRAC.Core.Utilities.File import getSize


__RCSID__ = "$Id$"


class ProxyStorage(StorageBase):

    _INPUT_PROTOCOLS = ["file", "dip", "dips"]
    _OUTPUT_PROTOCOLS = ["dip", "dips"]

    def __init__(self, storageName, parameters):

        StorageBase.__init__(self, storageName, parameters)
        self.pluginName = "Proxy"
        self.isok = True
        self.url = "DataManagement/StorageElementProxy"

    ######################################
    # File transfer functionalities
    ######################################

    def getFile(self, path, localPath=False):
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        failed = {}
        successful = {}
        client = RPCClient(self.url)
        # Make sure transferClient uses the same ProxyStorage instance.
        # Only the this one holds the file we want to transfer.
        transferClient = TransferClient(client.serviceURL)
        for src_url in urls.keys():
            res = client.prepareFile(self.name, src_url)
            if not res["OK"]:
                gLogger.error("ProxyStorage.getFile: Failed to prepare file on remote server.", res["Message"])
                failed[src_url] = res["Message"]
            else:
                fileName = os.path.basename(src_url)
                if localPath:
                    dest_file = "%s/%s" % (localPath, fileName)
                else:
                    dest_file = "%s/%s" % (os.getcwd(), fileName)
                res = transferClient.receiveFile(dest_file, "getFile/%s" % fileName)
                if not res["OK"]:
                    gLogger.error("ProxyStorage.getFile: Failed to recieve file from proxy server.", res["Message"])
                    failed[src_url] = res["Message"]
                elif not os.path.exists(dest_file):
                    errStr = "ProxyStorage.getFile: The destination local file does not exist."
                    gLogger.error(errStr, dest_file)
                    failed[src_url] = errStr
                else:
                    destSize = getSize(dest_file)
                    if destSize == -1:
                        errStr = "ProxyStorage.getFile: Failed to get the local file size."
                        gLogger.error(errStr, dest_file)
                        failed[src_url] = errStr
                    else:
                        successful[src_url] = destSize
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def putFile(self, path, sourceSize=0):

        client = RPCClient(self.url)

        if sourceSize:
            gLogger.debug(
                "ProxyStorage.putFile: The client has provided the source file size implying\
         a replication is requested."
            )
            return client.callProxyMethod(self.name, "putFile", [path], {"sourceSize": sourceSize})

        gLogger.debug("ProxyStorage.putFile: No source size was provided therefore a simple put will be performed.")
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        failed = {}
        successful = {}
        # make sure transferClient uses the same ProxyStorage instance we uploaded the file to
        transferClient = TransferClient(client.serviceURL)
        for dest_url, src_file in urls.items():
            fileName = os.path.basename(dest_url)
            res = transferClient.sendFile(src_file, "putFile/%s" % fileName)
            if not res["OK"]:
                gLogger.error("ProxyStorage.putFile: Failed to send file to proxy server.", res["Message"])
                failed[dest_url] = res["Message"]
            else:
                res = client.uploadFile(self.name, dest_url)
                if not res["OK"]:
                    gLogger.error(
                        "ProxyStorage.putFile: Failed to upload file to storage element from proxy server.",
                        res["Message"],
                    )
                    failed[dest_url] = res["Message"]
                else:
                    res = self.__executeOperation(dest_url, "getFileSize")
                    if not res["OK"]:
                        gLogger.error(
                            "ProxyStorage.putFile: Failed to determine destination file size.", res["Message"]
                        )
                        failed[dest_url] = res["Message"]
                    else:
                        successful[dest_url] = res["Value"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ######################################
    # File manipulation functionalities
    ######################################

    def exists(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "exists", [path], {})

    def isFile(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "isFile", [path], {})

    def getFileSize(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "getFileSize", [path], {})

    def getFileMetadata(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "getFileMetadata", [path], {})

    def removeFile(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "removeFile", [path], {})

    def prestageFile(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "prestageFile", [path], {})

    def prestageFileStatus(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "prestageFileStatus", [path], {})

    def pinFile(self, path, lifetime=60 * 60 * 24):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "pinFile", [path], {"lifetime": lifetime})

    def releaseFile(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "releaseFile", [path], {})

    ######################################
    # Directory manipulation functionalities
    ######################################

    def isDirectory(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "isDirectory", [path], {})

    def getDirectoryMetadata(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "getDirectoryMetadata", [path], {})

    def getDirectorySize(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "getDirectorySize", [path], {})

    def listDirectory(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "listDirectory", [path], {})

    def createDirectory(self, path):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "createDirectory", [path], {})

    def removeDirectory(self, path, recursive=False):
        client = RPCClient(self.url)
        return client.callProxyMethod(self.name, "removeDirectory", [path], {"recursive": recursive})

    def getDirectory(self, path):
        return S_ERROR("Not supported")

    def putDirectory(self, path):
        return S_ERROR("Not supported")

    def __executeOperation(self, url, method):
        """Executes the requested functionality with the supplied url"""
        fcn = None
        if hasattr(self, method) and callable(getattr(self, method)):
            fcn = getattr(self, method)
        if not fcn:
            return S_ERROR("Unable to invoke %s, it isn't a member function of ProxyStorage" % method)
        res = fcn([url])
        if not res["OK"]:
            return res
        elif url not in res["Value"]["Successful"]:
            return S_ERROR(res["Value"]["Failed"][url])
        return S_OK(res["Value"]["Successful"][url])



""" GOCDBClient module is a client for the GOC DB, looking for Downtimes.

    WARN: the URL of the GOC DB API is hardcoded, and is: https://goc.egi.eu/gocdbpi/public/
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


import six
import time
import socket
import requests

from datetime import datetime, timedelta
from xml.dom import minidom

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Security.Locations import getCAsLocation


def _parseSingleElement(element, attributes=None):
    """
    Given a DOM Element, return a dictionary of its child elements and values (as strings).
    """

    handler = {}
    for child in element.childNodes:
        attrName = str(child.nodeName)
        if attributes is not None:
            if attrName not in attributes:
                continue

        if child.nodeName == "SCOPES":
            for subchild in child.childNodes:
                if subchild.childNodes:
                    handler.setdefault("SCOPES", []).append(subchild.childNodes[0].nodeValue)
            continue

        if child.nodeName == "EXTENSIONS":
            for subchild in child.childNodes:
                if subchild.childNodes:
                    dct = {}
                    for subsubchild in subchild.childNodes:
                        if subsubchild.childNodes:
                            dct[subsubchild.nodeName] = subsubchild.childNodes[0].nodeValue
                    handler.setdefault("EXTENSIONS", []).append(dct)
            continue

        try:
            attrValue = child.childNodes[0].nodeValue
        except IndexError:
            continue
        handler[attrName] = attrValue

    return handler


#############################################################################


class GOCDBClient(object):
    """Class for dealing with GOCDB. Class because of easier use from RSS"""

    #############################################################################

    def getStatus(self, granularity, name=None, startDate=None, startingInHours=None, timeout=None):
        """
        Return actual GOCDB status of entity in `name`

        :param str granularity: should be a ValidRes, e.g. "Resource"
        :param name: should be the name(s) of the ValidRes.
          Could be a list of strings or simply one string.
          If not given, fetches the complete list.

        :param startDate: if not given, takes only ongoing DownTimes.
          if given, could be a datetime or a string ("YYYY-MM-DD"), and download
          DownTimes starting after that date.

        :param int startingInHours: optional integer. If given, donwload
          DownTimes starting in the next given hours (startDate is then useless)

        :return: (example)

          .. code-block:: python

            {'OK': True,
             'Value': {'92569G0 lhcbsrm-kit.gridka.de': {
                           'DESCRIPTION': 'Annual site downtime for various major tasks i...',
                           'FORMATED_END_DATE': '2014-05-27 15:21',
                           'FORMATED_START_DATE': '2014-05-26 04:00',
                           'GOCDB_PORTAL_URL': 'https://goc.egi.eu/portal/index.php?Page_Type=Downtime&id=14051',
                           'HOSTED_BY': 'FZK-LCG2',
                           'HOSTNAME': 'lhcbsrm-kit.gridka.de',
                           'SERVICE_TYPE': 'SRM.nearline',
                           'SEVERITY': 'OUTAGE'},
                       '99873G0 srm.pic.esSRM': {
                            'HOSTED_BY': 'pic',
                            'ENDPOINT': 'srm.pic.esSRM',
                            'SEVERITY': 'OUTAGE',
                            'HOSTNAME': 'srm.pic.es',
                            'URL': 'srm.pic.es',
                            'GOCDB_PORTAL_URL': 'https://goc.egi.eu/portal/index.php?Page_Type=Downtime&id=21303',
                            'FORMATED_START_DATE': '2016-09-14 06:00',
                            'SERVICE_TYPE': 'SRM',
                            'FORMATED_END_DATE': '2016-09-14 15:00',
                            'DESCRIPTION': 'Outage declared due to network and dCache upgrades'}
                       }
              }
        """

        startDate_STR = None
        startDateMax = None

        if startingInHours is not None:
            startDate = datetime.utcnow()
            startDateMax = startDate + timedelta(hours=startingInHours)

        if startDate is not None:
            if isinstance(startDate, six.string_types):
                startDate_STR = startDate
                startDate = datetime(*time.strptime(startDate, "%Y-%m-%d")[0:3])
            elif isinstance(startDate, datetime):
                startDate_STR = startDate.isoformat(" ")[0:10]

        if timeout is not None:
            socket.setdefaulttimeout(10)

        if startingInHours is not None:
            # make 2 queries and later merge the results

            # first call: pass the startDate argument as None,
            # so the curlDownload method will search for only ongoing DTs
            resXML_ongoing = self._downTimeCurlDownload(name)
            if resXML_ongoing is None:
                resOngoing = {}
            else:
                resOngoing = self._downTimeXMLParsing(resXML_ongoing, granularity, name)

            # second call: pass the startDate argument
            resXML_startDate = self._downTimeCurlDownload(name, startDate_STR)
            if resXML_startDate is None:
                resStartDate = {}
            else:
                resStartDate = self._downTimeXMLParsing(resXML_startDate, granularity, name, startDateMax)

            # merge the results of the 2 queries:
            res = resOngoing
            for k in resStartDate:
                if k not in res:
                    res[k] = resStartDate[k]

        else:
            # just query for onGoing downtimes
            resXML = self._downTimeCurlDownload(name, startDate_STR)
            if resXML is None:
                return S_OK(None)

            res = self._downTimeXMLParsing(resXML, granularity, name, startDateMax)

        if res == {}:
            res = None

        return S_OK(res)

    def getServiceEndpointInfo(self, granularity, entity):
        """
        Get service endpoint info (in a dictionary)

        :params:
          :attr:`granularity` : a string. Could be in ('hostname', 'sitename', 'roc',
          'country', 'service_type', 'monitored')

          :attr:`entity` : a string. Actual name of the entity.
        """
        assert isinstance(granularity, six.string_types) and isinstance(entity, six.string_types)
        try:
            serviceXML = self._getServiceEndpointCurlDownload(granularity, entity)
            return S_OK(self._serviceEndpointXMLParsing(serviceXML))
        except Exception as e:
            _msg = "Exception getting information for %s %s: %s" % (granularity, entity, e)
            gLogger.exception(_msg)
            return S_ERROR(_msg)

    def getCurrentDTLinkList(self):
        """
        Get the list of all current DTs' links
        """

        gDTPage = self._downTimeCurlDownload()  # xml format
        gResourceDT = self._downTimeXMLParsing(gDTPage, "Resource")  # python dictionary format
        gSiteDT = self._downTimeXMLParsing(gDTPage, "Site")  # python dictionary format

        currentDTLinkList = []
        for dt in gResourceDT:
            currentDTLinkList.append(gResourceDT[dt]["GOCDB_PORTAL_URL"])

        for dt in gSiteDT:
            currentDTLinkList.append(gSiteDT[dt]["GOCDB_PORTAL_URL"])

        return S_OK(currentDTLinkList)

    #############################################################################

    def getHostnameDowntime(self, hostname, startDate=None, ongoing=False):

        params = hostname

        if startDate and ongoing:
            return S_ERROR("Invalid parameter combination - do not specify startDate with ongoing")

        if startDate:
            params += "&startdate=" + startDate

        if ongoing:
            params += "&ongoing_only=yes"

        caPath = getCAsLocation()

        try:
            response = requests.get(
                "https://goc.egi.eu/gocdbpi/public/?method=get_downtime&topentity=" + params, verify=caPath
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            return S_ERROR("Error %s" % e)

        return S_OK(response.text)

    #############################################################################

    #  def getSiteInfo(self, site):
    #    """
    #    Get site info (in a dictionary)
    #
    #    :params:
    #      :attr:`entity` : a string. Actual name of the site.
    #    """
    #
    #    siteXML = self._getSiteCurlDownload(site)
    #    return S_OK(self._siteXMLParsing(siteXML))

    #############################################################################

    #  def buildURL(self, DTList):
    #    '''build the URL relative to the DT '''
    #    baseURL = "https://goc.egi.eu/downtime/list?id="
    #    for dt in DTList:
    #      id = str(dt['id'])
    #      url = baseURL + id
    #      dt['URL'] = url

    #############################################################################

    def _downTimeCurlDownload(self, entity=None, startDate=None):
        """Download ongoing downtimes for entity using the GOC DB programmatic interface"""

        # GOCDB-PI url and method settings
        #
        # Set the GOCDB URL
        gocdbpi_url = "https://goc.egi.eu/gocdbpi/public/?method=get_downtime"
        # Set the desidered start date
        if startDate is None:
            when = "&ongoing_only=yes"
            gocdbpi_startDate = ""
        else:
            when = "&startdate="
            gocdbpi_startDate = startDate

        # GOCDB-PI to query
        gocdb_ep = gocdbpi_url
        if entity is not None:
            if isinstance(entity, six.string_types):
                gocdb_ep = gocdb_ep + "&topentity=" + entity
        gocdb_ep = gocdb_ep + when + gocdbpi_startDate + "&scope="

        caPath = getCAsLocation()
        dtPage = requests.get(gocdb_ep, verify=caPath)

        dt = dtPage.text

        return dt

    #############################################################################

    def _getServiceEndpointCurlDownload(self, granularity, entity):
        """
        Calls method `get_service_endpoint` from the GOC DB programmatic interface.

        :params:
          :attr:`granularity` : a string. Could be in ('hostname', 'sitename', 'roc',
          'country', 'service_type', 'monitored')

          :attr:`entity` : a string. Actual name of the entity.
        """
        if not isinstance(granularity, six.string_types) or not isinstance(entity, six.string_types):
            raise ValueError("Arguments must be strings.")

        # GOCDB-PI query
        gocdb_ep = "https://goc.egi.eu/gocdbpi/public/?method=get_service_endpoint&" + granularity + "=" + entity

        caPath = getCAsLocation()
        service_endpoint_page = requests.get(gocdb_ep, verify=caPath)

        return service_endpoint_page.text

    #############################################################################

    #  def _getSiteCurlDownload(self, site):
    #    """
    #    Calls method `get_site` from the GOC DB programmatic interface.
    #
    #    :params:
    #      :attr:`site` : a string. Actual name of the site.
    #    """
    #
    #    # GOCDB-PI query
    #    gocdb_ep = "https://goc.egi.eu/gocdbpi/public/?method=get_site&sitename="+site
    #
    #    site_page = requests.get( gocdb_ep )
    #
    #    return site_page.text

    #############################################################################

    def _downTimeXMLParsing(self, dt, siteOrRes, entities=None, startDateMax=None):
        """Performs xml parsing from the dt string (returns a dictionary)"""
        doc = minidom.parseString(dt)

        downtimeElements = doc.getElementsByTagName("DOWNTIME")
        dtDict = {}

        for dtElement in downtimeElements:
            elements = _parseSingleElement(
                dtElement,
                [
                    "SEVERITY",
                    "SITENAME",
                    "HOSTNAME",
                    "ENDPOINT",
                    "HOSTED_BY",
                    "FORMATED_START_DATE",
                    "FORMATED_END_DATE",
                    "DESCRIPTION",
                    "GOCDB_PORTAL_URL",
                    "SERVICE_TYPE",
                ],
            )
            try:
                affectedEndpoints = dtElement.getElementsByTagName("AFFECTED_ENDPOINTS")
                urls = []
                for epElement in affectedEndpoints[0].childNodes:
                    try:
                        urls.append(_parseSingleElement(epElement, ["URL"])["URL"])
                    except (IndexError, KeyError):
                        pass
            except (IndexError, KeyError):
                pass

            try:
                dtDict[str(dtElement.getAttributeNode("PRIMARY_KEY").nodeValue) + " " + elements["ENDPOINT"]] = elements
                dtDict[str(dtElement.getAttributeNode("PRIMARY_KEY").nodeValue) + " " + elements["ENDPOINT"]][
                    "URL"
                ] = urls[0]
            except Exception:
                try:
                    dtDict[
                        str(dtElement.getAttributeNode("PRIMARY_KEY").nodeValue) + " " + elements["HOSTNAME"]
                    ] = elements
                except Exception:
                    dtDict[
                        str(dtElement.getAttributeNode("PRIMARY_KEY").nodeValue) + " " + elements["SITENAME"]
                    ] = elements

        for dtID in list(dtDict):
            if siteOrRes in ("Site", "Sites"):
                if "SITENAME" not in dtDict[dtID]:
                    dtDict.pop(dtID)
                    continue
                if entities is not None:
                    if not isinstance(entities, list):
                        entities = [entities]
                    if not dtDict[dtID]["SITENAME"] in entities:
                        dtDict.pop(dtID)

            elif siteOrRes in ("Resource", "Resources"):
                if "HOSTNAME" not in dtDict[dtID]:
                    dtDict.pop(dtID)
                    continue
                if entities is not None:
                    if not isinstance(entities, list):
                        entities = [entities]
                    hostname = dtDict[dtID].get("URL", dtDict[dtID]["HOSTNAME"])
                    if hostname not in entities:
                        dtDict.pop(dtID)

        if startDateMax is not None:
            for dtID in list(dtDict):
                startDateMaxFromKeys = datetime(
                    *time.strptime(dtDict[dtID]["FORMATED_START_DATE"], "%Y-%m-%d %H:%M")[0:5]
                )
                if startDateMaxFromKeys > startDateMax:
                    dtDict.pop(dtID)

        return dtDict

    def _serviceEndpointXMLParsing(self, serviceXML):
        """Performs xml parsing from the service endpoint string
        Returns a list.
        """
        doc = minidom.parseString(serviceXML)
        services = doc.getElementsByTagName("SERVICE_ENDPOINT")
        services = [_parseSingleElement(s) for s in services]
        return services



""" :mod: GFAL2_SRM2Storage

    =================

    .. module: python

    :synopsis: SRM2 module based on the GFAL2_StorageBase class.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=invalid-name

import six
import errno
import json

import gfal2  # pylint: disable=import-error

# from DIRAC
from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Resources.Storage.GFAL2_StorageBase import GFAL2_StorageBase
from DIRAC.Resources.Storage.Utilities import checkArgumentFormat


__RCSID__ = "$Id$"


class GFAL2_SRM2Storage(GFAL2_StorageBase):
    """SRM2 SE class that inherits from GFAL2StorageBase"""

    _INPUT_PROTOCOLS = ["file", "root", "srm", "gsiftp"]
    _OUTPUT_PROTOCOLS = ["file", "root", "dcap", "gsidcap", "rfio", "srm", "gsiftp"]

    def __init__(self, storageName, parameters):
        """ """
        super(GFAL2_SRM2Storage, self).__init__(storageName, parameters)
        self.log = gLogger.getSubLogger("GFAL2_SRM2Storage")
        self.log.debug("GFAL2_SRM2Storage.__init__: Initializing object")
        self.pluginName = "GFAL2_SRM2"

        # This attribute is used to know the file status (OFFLINE,NEARLINE,ONLINE)
        self._defaultExtendedAttributes = ["user.status"]

        # ##
        #    Setting the default SRM parameters here. For methods where this
        #    is not the default there is a method defined in this class, setting
        #    the proper values and then calling the base class method.
        # ##

        self.gfal2requestLifetime = gConfig.getValue("/Resources/StorageElements/RequestLifeTime", 100)

        self.__setSRMOptionsToDefault()

        # This lists contains the list of protocols to ask to SRM to get a URL
        # It can be either defined in the plugin of the SE, or as a global option
        if "ProtocolsList" in parameters:
            self.protocolsList = parameters["ProtocolsList"].split(",")
        else:
            self.log.debug("GFAL2_SRM2Storage: No protocols provided, using the default protocols.")
            self.protocolsList = self.defaultLocalProtocols
            self.log.debug("GFAL2_SRM2Storage: protocolsList = %s" % self.protocolsList)

    def __setSRMOptionsToDefault(self):
        """Resetting the SRM options back to default"""
        self.ctx.set_opt_integer("SRM PLUGIN", "OPERATION_TIMEOUT", self.gfal2Timeout)
        if self.spaceToken:
            self.ctx.set_opt_string("SRM PLUGIN", "SPACETOKENDESC", self.spaceToken)
        self.ctx.set_opt_integer("SRM PLUGIN", "REQUEST_LIFETIME", self.gfal2requestLifetime)
        # Setting the TURL protocol to gsiftp because with other protocols we have authorisation problems
        #    self.ctx.set_opt_string_list( "SRM PLUGIN", "TURL_PROTOCOLS", self.defaultLocalProtocols )
        self.ctx.set_opt_string_list("SRM PLUGIN", "TURL_PROTOCOLS", ["gsiftp"])

    def _updateMetadataDict(self, metadataDict, attributeDict):
        """Updating the metadata dictionary with srm specific attributes

        :param self: self reference
        :param dict: metadataDict we want add the SRM specific attributes to
        :param dict: attributeDict contains 'user.status' which we then fill in the metadataDict

        """
        # 'user.status' is the extended attribute we are interested in
        user_status = attributeDict.get("user.status", "")
        metadataDict["Cached"] = int("ONLINE" in user_status)
        metadataDict["Migrated"] = int("NEARLINE" in user_status)
        metadataDict["Lost"] = int(user_status == "LOST")
        metadataDict["Unavailable"] = int(user_status == "UNAVAILABLE")
        metadataDict["Accessible"] = (
            not metadataDict["Lost"] and metadataDict["Cached"] and not metadataDict["Unavailable"]
        )

    def getTransportURL(self, path, protocols=False):
        """obtain the tURLs for the supplied path and protocols

        :param self: self reference
        :param str path: path on storage
        :param mixed protocols: protocols to use
        :returns: Failed dict {path : error message}
                 Successful dict {path : transport url}
                 S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_SRM2Storage.getTransportURL: Attempting to retrieve tURL for %s paths" % len(urls))

        failed = {}
        successful = {}
        if not protocols:
            listProtocols = self.protocolsList
            if not listProtocols:
                return S_ERROR("GFAL2_SRM2Storage.getTransportURL: No local protocols defined and no defaults found.")
        elif isinstance(protocols, six.string_types):
            listProtocols = [protocols]
        elif isinstance(protocols, list):
            listProtocols = protocols
        else:
            return S_ERROR("getTransportURL: Must supply desired protocols to this plug-in.")

        # Compatibility because of castor returning a castor: url if you ask
        # for a root URL, and a root: url if you ask for a xroot url...
        if "root" in listProtocols and "xroot" not in listProtocols:
            listProtocols.insert(listProtocols.index("root"), "xroot")
        elif "xroot" in listProtocols and "root" not in listProtocols:
            listProtocols.insert(listProtocols.index("xroot") + 1, "root")

        if self.protocolParameters["Protocol"] in listProtocols:
            successful = {}
            failed = {}
            for url in urls:
                if self.isURL(url)["Value"]:
                    successful[url] = url
                else:
                    failed[url] = "getTransportURL: Failed to obtain turls."

            return S_OK({"Successful": successful, "Failed": failed})

        for url in urls:
            res = self.__getSingleTransportURL(url, listProtocols)
            self.log.debug("res = %s" % res)

            if not res["OK"]:
                failed[url] = res["Message"]
            else:
                successful[url] = res["Value"]

        return S_OK({"Failed": failed, "Successful": successful})

    def __getSingleTransportURL(self, path, protocols=False):
        """Get the tURL from path with getxattr from gfal2

        :param self: self reference
        :param str path: path on the storage
        :returns: S_OK( Transport_URL ) in case of success
                  S_ERROR( errStr ) in case of a failure
        """
        self.log.debug("GFAL2_SRM2Storage.__getSingleTransportURL: trying to retrieve tURL for %s" % path)
        if protocols:
            self.ctx.set_opt_string_list("SRM PLUGIN", "TURL_PROTOCOLS", protocols)

        res = self._getExtendedAttributes(path, attributes=["user.replicas"])
        self.__setSRMOptionsToDefault()

        if res["OK"]:
            return S_OK(res["Value"]["user.replicas"])

        errStr = "GFAL2_SRM2Storage.__getSingleTransportURL: Extended attribute tURL is not set."
        self.log.debug(errStr, res["Message"])
        return res

    def getOccupancy(self, *parms, **kws):
        """Gets the GFAL2_SRM2Storage occupancy info.

        TODO: needs gfal2.15 because of bugs:
        https://its.cern.ch/jira/browse/DMC-979
        https://its.cern.ch/jira/browse/DMC-977

        It queries the srm interface for a given space token.
        Out of the results, we keep totalsize, guaranteedsize, and unusedsize all in B.
        """

        if not self.spaceToken:
            self.log.info("getOccupancy: SpaceToken not defined for this SE. Falling back to the default getOccupancy.")
            return super(GFAL2_SRM2Storage, self).getOccupancy(*parms, **kws)

        # Gfal2 extended parameter name to query the space token occupancy
        spaceTokenAttr = "spacetoken.description?%s" % self.protocolParameters["SpaceToken"]
        # gfal2 can take any srm url as a base.
        spaceTokenEndpoint = self.getURLBase(withWSUrl=True)["Value"]
        try:
            occupancyStr = self.ctx.getxattr(spaceTokenEndpoint, spaceTokenAttr)
            try:
                occupancyDict = json.loads(occupancyStr)[0]
            except ValueError:
                # https://its.cern.ch/jira/browse/DMC-977
                # a closing bracket is missing, so we retry after adding it
                occupancyStr = occupancyStr[:-1] + "}]"
                occupancyDict = json.loads(occupancyStr)[0]

                # https://its.cern.ch/jira/browse/DMC-979
                # We set totalsize to guaranteed size
                # (it is anyway true for all the SEs I could test)
                occupancyDict["totalsize"] = occupancyDict.get("guaranteedsize", 0)

        except (gfal2.GError, ValueError) as e:
            errStr = "Something went wrong while checking for spacetoken occupancy."
            self.log.verbose(errStr, e.message)
            return S_ERROR(getattr(e, "code", errno.EINVAL), "%s %s" % (errStr, repr(e)))

        sTokenDict = {}

        sTokenDict["Total"] = float(occupancyDict.get("totalsize", "0"))
        sTokenDict["Free"] = float(occupancyDict.get("unusedsize", "0"))
        sTokenDict["SpaceReservation"] = self.protocolParameters["SpaceToken"]

        return S_OK(sTokenDict)



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-get-CAs
# Author :  Ricardo Graciani
########################################################################
"""
Refresh the local copy of the CA certificates and revocation lists.

Connects to the BundleDelivery service to obtain the tar balls. Needed when proxies appear to be
invalid.

Example:
  $ dirac-admin-get-CAs
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.FrameworkSystem.Client.BundleDeliveryClient import BundleDeliveryClient

__RCSID__ = "$Id$"


@Script()
def main():
    Script.addDefaultOptionValue("/DIRAC/Security/SkipCAChecks", "yes")
    Script.parseCommandLine(ignoreErrors=True)

    bdc = BundleDeliveryClient()

    result = bdc.syncCAs()
    if not result["OK"]:
        DIRAC.gLogger.error("Error while updating CAs", result["Message"])
        DIRAC.exit(1)
    elif result["Value"]:
        DIRAC.gLogger.notice("CAs got updated")
    else:
        DIRAC.gLogger.notice("CAs are already synchronized")

    result = bdc.syncCRLs()
    if not result["OK"]:
        DIRAC.gLogger.error("Error while updating CRLs", result["Message"])
        DIRAC.exit(1)
    elif result["Value"]:
        DIRAC.gLogger.notice("CRLs got updated")
    else:
        DIRAC.gLogger.notice("CRLs are already synchronized")

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" DIRAC DirectoryTree base class """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import time
import threading
import os
import stat

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.Utilities import getIDSelectString

DEBUG = 0

#############################################################################


class DirectoryTreeBase(object):
    def __init__(self, database=None):
        self.db = database
        self.lock = threading.Lock()
        self.treeTable = ""

    ############################################################################
    #
    # THE FOLLOWING METHODS NEED TO ME IMPLEMENTED IN THE DERIVED CLASS
    #
    ############################################################################

    def findDir(self, path, connection=False):
        """Find directory ID for the given path"""
        return S_ERROR("To be implemented on derived class")

    def findDirs(self, paths, connection=False):
        """Find DirIDs for the given path list"""
        return S_ERROR("To be implemented on derived class")

    def makeDir(self, path):

        return S_ERROR("To be implemented on derived class")

    def removeDir(self, path):

        return S_ERROR("To be implemented on derived class")

    def getChildren(self, path, connection=False):
        return S_ERROR("To be implemented on derived class")

    def getDirectoryPath(self, dirID):
        """Get directory name by directory ID"""
        return S_ERROR("To be implemented on derived class")

    def countSubdirectories(self, dirId, includeParent=True):
        return S_ERROR("To be implemented on derived class")

    def getSubdirectoriesByID(self, dirID, requestString=False, includeParent=False):
        """Get all the subdirectories of the given directory at a given level"""
        return S_ERROR("To be implemented on derived class")

    ##########################################################################

    def _getConnection(self, connection):
        if connection:
            return connection
        res = self.db._getConnection()
        if res["OK"]:
            return res["Value"]
        gLogger.warn("Failed to get MySQL connection", res["Message"])
        return connection

    def getTreeTable(self):
        """Get the string of the Directory Tree type"""
        return self.treeTable

    def setDatabase(self, database):
        self.db = database

    def makeDirectory(self, path, credDict, status=0):
        """Create a new directory. The return value is the dictionary
        containing all the parameters of the newly created directory
        """
        if path[0] != "/":
            return S_ERROR("Not an absolute path")
        # Strip off the trailing slash if necessary
        if len(path) > 1 and path[-1] == "/":
            path = path[:-1]

        if path == "/":
            # Create the root directory
            l_uid = 0
            l_gid = 0
        else:
            result = self.db.ugManager.getUserAndGroupID(credDict)
            if not result["OK"]:
                return result
            (l_uid, l_gid) = result["Value"]

        dirDict = {}
        result = self.makeDir(path)
        if not result["OK"]:
            return result
        dirID = result["Value"]
        if result["NewDirectory"]:
            req = "INSERT INTO FC_DirectoryInfo (DirID,UID,GID,CreationDate,ModificationDate,Mode,Status) Values "
            req = req + "(%d,%d,%d,UTC_TIMESTAMP(),UTC_TIMESTAMP(),%d,%d)" % (
                dirID,
                l_uid,
                l_gid,
                self.db.umask,
                status,
            )
            result = self.db._update(req)
            if result["OK"]:
                resGet = self.getDirectoryParameters(dirID)
                if resGet["OK"]:
                    dirDict = resGet["Value"]
        else:
            return S_OK(dirID)

        if not dirDict:
            self.removeDir(path)
            return S_ERROR("Failed to create directory %s" % path)
        return S_OK(dirID)

    #####################################################################
    def makeDirectories(self, path, credDict):
        """Make all the directories recursively in the path. The return value
        is the dictionary containing all the parameters of the newly created
        directory
        """

        if not path or path[0] != "/":
            return S_ERROR("Not an absolute path")

        result = self.existsDir(path)
        if not result["OK"]:
            return result
        result = result["Value"]
        if result["Exists"]:
            return S_OK(result["DirID"])

        if path == "/":
            result = self.makeDirectory(path, credDict)
            return result

        parentDir = os.path.dirname(path)
        result = self.existsDir(parentDir)
        if not result["OK"]:
            return result
        result = result["Value"]
        if result["Exists"]:
            result = self.makeDirectory(path, credDict)
        else:
            result = self.makeDirectories(parentDir, credDict)
            if not result["OK"]:
                return result
            result = self.makeDirectory(path, credDict)

        return result

    #####################################################################
    def exists(self, lfns):
        successful = {}
        failed = {}
        for lfn in lfns:
            res = self.findDir(lfn)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            if not res["Value"]:
                successful[lfn] = False
            else:
                successful[lfn] = lfn
        return S_OK({"Successful": successful, "Failed": failed})

    def existsDir(self, path):
        """Check the existence of the directory path"""
        result = self.findDir(path)
        if not result["OK"]:
            return result
        if result["Value"]:
            result = S_OK(int(result["Value"]))
            result["Exists"] = True
            result["DirID"] = result["Value"]
        else:
            result = S_OK(0)
            result["Exists"] = False

        return result

    #####################################################################
    def isDirectory(self, paths):
        """Checking for existence of directories"""
        successful = {}
        failed = {}
        for dir in paths:
            result = self.existsDir(dir)
            if not result["OK"]:
                failed[dir] = result["Message"]
            elif result["Value"]["Exists"]:
                successful[dir] = True
            else:
                successful[dir] = False

        return S_OK({"Successful": successful, "Failed": failed})

    #####################################################################
    def createDirectory(self, dirs, credDict):
        """Checking for existence of directories"""
        successful = {}
        failed = {}
        for dir in dirs:
            result = self.makeDirectories(dir, credDict)
            if not result["OK"]:
                failed[dir] = result["Message"]
            else:
                successful[dir] = True

        return S_OK({"Successful": successful, "Failed": failed})

    #####################################################################
    def isEmpty(self, path):
        """Find out if the given directory is empty"""
        # Check if there are subdirectories
        result = self.getChildren(path)
        if not result["OK"]:
            return result
        childIDs = result["Value"]
        if childIDs:
            return S_OK(False)

        # Check if there are files
        result = self.__getDirID(path)
        if not result["OK"]:
            return result
        dirID = result["Value"]
        result = self.db.fileManager.getFilesInDirectory(dirID)
        if not result["OK"]:
            return result
        files = result["Value"]
        if files:
            return S_OK(False)

        return S_OK(True)

    #####################################################################
    def removeDirectory(self, dirs, force=False):
        """Remove an empty directory from the catalog"""
        successful = {}
        failed = {}

        # Check if requested directories exist in the catalog
        result = self.findDirs(dirs)
        if not result["OK"]:
            return result
        dirDict = result["Value"]
        for d in dirs:
            if d not in dirDict:
                successful[d] = "Directory does not exist"

        for dir in dirDict:
            result = self.isEmpty(dir)
            if not result["OK"]:
                return result
            if not result["Value"]:
                failed[dir] = "Failed to remove non-empty directory"
                continue
            result = self.removeDir(dir)
            if not result["OK"]:
                failed[dir] = result["Message"]
            else:
                successful[dir] = result
        return S_OK({"Successful": successful, "Failed": failed})

    #####################################################################
    def __getDirID(self, path):
        """Get directory ID from the given path or already evaluated ID"""

        if isinstance(path, six.string_types):
            result = self.findDir(path)
            if not result["OK"]:
                return result
            dirID = result["Value"]
            if not dirID:
                return S_ERROR("%s: not found" % str(path))
            return S_OK(dirID)
        else:
            return S_OK(path)

    #####################################################################
    def getDirectoryParameters(self, path):
        """Get the given directory parameters"""

        result = self.__getDirID(path)
        if not result["OK"]:
            return result
        dirID = result["Value"]

        query = "SELECT DirID,UID,GID,Status,Mode,CreationDate,ModificationDate from FC_DirectoryInfo"
        query = query + " WHERE DirID=%d" % dirID
        resQuery = self.db._query(query)
        if not resQuery["OK"]:
            return resQuery

        if not resQuery["Value"]:
            return S_ERROR("Directory not found")
        dirDict = {}
        dirDict["DirID"] = int(resQuery["Value"][0][0])
        uid = int(resQuery["Value"][0][1])
        dirDict["UID"] = uid
        owner = "unknown"
        result = self.db.ugManager.getUserName(uid)
        if result["OK"]:
            owner = result["Value"]
        dirDict["Owner"] = owner
        gid = int(resQuery["Value"][0][2])
        dirDict["GID"] = int(resQuery["Value"][0][2])
        group = "unknown"
        result = self.db.ugManager.getGroupName(gid)
        if result["OK"]:
            group = result["Value"]
        dirDict["OwnerGroup"] = group
        dirDict["Status"] = int(resQuery["Value"][0][3])
        dirDict["Mode"] = int(resQuery["Value"][0][4])
        dirDict["CreationDate"] = resQuery["Value"][0][5]
        dirDict["ModificationDate"] = resQuery["Value"][0][6]

        return S_OK(dirDict)

    #####################################################################
    def _setDirectoryParameter(self, path, pname, pvalue):
        """Set a numerical directory parameter

        :param mixed path: Directory path or paths as a string or directory ID as int,
                           list/tuple of ints or a string to select directory IDs
        :param str pname: parameter name
        :param int pvalue: parameter value
        """
        result = getIDSelectString(path)
        if not result["OK"] and isinstance(path, six.string_types):
            result = self.__getDirID(path)
            if not result["OK"]:
                return result
            dirID = result["Value"]
            result = getIDSelectString(dirID)
            if not result["OK"]:
                return result

        dirIDString = result["Value"]
        req = "UPDATE FC_DirectoryInfo SET %s=%d, " "ModificationDate=UTC_TIMESTAMP() WHERE DirID IN ( %s )" % (
            pname,
            pvalue,
            dirIDString,
        )
        result = self.db._update(req)
        return result

    #####################################################################
    def _setDirectoryGroup(self, path, gname):
        """Set the directory group

        :param mixed path: directory path as a string or int or list of ints or select statement
        :param mixt group: new group as a string or int gid
        """

        result = self.db.ugManager.findGroup(gname)
        if not result["OK"]:
            return result

        gid = result["Value"]

        return self._setDirectoryParameter(path, "GID", gid)

    #####################################################################
    def _setDirectoryOwner(self, path, owner):
        """Set the directory owner

        :param mixed path: directory path as a string or int or list of ints or select statement
        :param mixt owner: new user as a string or int uid
        """

        result = self.db.ugManager.findUser(owner)
        if not result["OK"]:
            return result

        uid = result["Value"]

        return self._setDirectoryParameter(path, "UID", uid)

    #####################################################################
    def changeDirectoryOwner(self, paths, recursive=False):
        """Bulk setting of the directory owner

        :param dictionary paths: dictionary < lfn : owner >
        """
        return self._changeDirectoryParameter(
            paths, self._setDirectoryOwner, self.db.fileManager.setFileOwner, recursive=recursive
        )

    #####################################################################
    def changeDirectoryGroup(self, paths, recursive=False):
        """Bulk setting of the directory group

        :param dictionary paths: dictionary < lfn : group >
        """
        return self._changeDirectoryParameter(
            paths, self._setDirectoryGroup, self.db.fileManager.setFileGroup, recursive=recursive
        )

    #####################################################################
    def _setDirectoryMode(self, path, mode):
        """set the directory mode

        :param mixed path: directory path as a string or int or list of ints or select statement
        :param int mode: new mode
        """
        return self._setDirectoryParameter(path, "Mode", mode)

    #####################################################################
    def changeDirectoryMode(self, paths, recursive=False):
        """Bulk setting of the directory mode

        :param dictionary paths: dictionary < lfn : mode >
        """
        return self._changeDirectoryParameter(
            paths, self._setDirectoryMode, self.db.fileManager.setFileMode, recursive=recursive
        )

    #####################################################################
    def _changeDirectoryParameter(self, paths, directoryFunction, fileFunction, recursive=False):
        """Bulk setting of the directory parameter with recursion for all the subdirectories and files

        :param dictionary paths: dictionary < lfn : value >, where value is the value of parameter to be set
        :param function directoryFunction: function to change directory(ies) parameter
        :param function fileFunction: function to change file(s) parameter
        :param bool recursive: flag to apply the operation recursively
        """
        arguments = paths
        successful = {}
        failed = {}
        for path, attribute in arguments.items():
            result = directoryFunction(path, attribute)
            if not result["OK"]:
                failed[path] = result["Message"]
                continue
            if recursive:
                result = self.__getDirID(path)
                if not result["OK"]:
                    failed[path] = result["Message"]
                    continue
                dirID = result["Value"]
                result = self.getSubdirectoriesByID(dirID, requestString=True, includeParent=True)
                if not result["OK"]:
                    failed[path] = result["Message"]
                    continue

                subDirQuery = result["Value"]
                result = self.db.fileManager.getFileIDsInDirectory(subDirQuery, requestString=True)
                if not result["OK"]:
                    failed[path] = result["Message"]
                    continue
                fileQuery = result["Value"]

                result = directoryFunction(subDirQuery, attribute)
                if not result["OK"]:
                    failed[path] = result["Message"]
                    continue
                result = fileFunction(fileQuery, attribute)
                if not result["OK"]:
                    failed[path] = result["Message"]
                else:
                    successful[path] = True
            else:
                successful[path] = True

        return S_OK({"Successful": successful, "Failed": failed})

    #####################################################################
    def setDirectoryStatus(self, path, status):
        """set the directory status"""
        return self._setDirectoryParameter(path, "Status", status)

    def getPathPermissions(self, lfns, credDict):
        """Get permissions for the given user/group to manipulate the given lfns"""
        successful = {}
        failed = {}
        for path in lfns:
            result = self.getDirectoryPermissions(path, credDict)
            if not result["OK"]:
                failed[path] = result["Message"]
            else:
                successful[path] = result["Value"]

        return S_OK({"Successful": successful, "Failed": failed})

    #####################################################################
    def getDirectoryPermissions(self, path, credDict):
        """Get permissions for the given user/group to manipulate the given directory"""
        result = self.db.ugManager.getUserAndGroupID(credDict)
        if not result["OK"]:
            return result
        uid, gid = result["Value"]

        result = self.getDirectoryParameters(path)
        if not result["OK"]:
            if "not found" in result["Message"] or "not exist" in result["Message"]:
                # If the directory does not exist, check the nearest parent for the permissions
                if path == "/":
                    # Nothing yet exists, starting from the scratch
                    resultDict = {}
                    resultDict["Write"] = True
                    resultDict["Read"] = True
                    resultDict["Execute"] = True
                    return S_OK(resultDict)
                else:
                    pDir = os.path.dirname(path)
                    if not pDir:
                        return S_ERROR("Illegal Path")
                    if pDir == path:
                        # If pDir == path, then we're stuck in a loop
                        # There is probably a "//" in the path
                        return S_ERROR("Bad Path (double /?)")
                    result = self.getDirectoryPermissions(pDir, credDict)
                    return result
            else:
                return result

        dUid = result["Value"]["UID"]
        dGid = result["Value"]["GID"]
        mode = result["Value"]["Mode"]

        owner = uid == dUid
        group = gid == dGid

        resultDict = {}
        if self.db.globalReadAccess:
            resultDict["Read"] = True
        else:
            resultDict["Read"] = (
                (owner and mode & stat.S_IRUSR > 0) or (group and mode & stat.S_IRGRP > 0) or mode & stat.S_IROTH > 0
            )

        resultDict["Write"] = (
            (owner and mode & stat.S_IWUSR > 0) or (group and mode & stat.S_IWGRP > 0) or mode & stat.S_IWOTH > 0
        )

        resultDict["Execute"] = (
            (owner and mode & stat.S_IXUSR > 0) or (group and mode & stat.S_IXGRP > 0) or mode & stat.S_IXOTH > 0
        )

        return S_OK(resultDict)

    def getFileIDsInDirectoryWithLimits(self, dirID, credDict, startItem=1, maxItems=25):
        """Get file IDs for the given directory"""
        dirs = dirID
        if not isinstance(dirID, list):
            dirs = [dirID]

        if not dirs:
            dirs = [-1]

        dirListString = ",".join([str(dir) for dir in dirs])

        req = "SELECT COUNT( DirID ) FROM FC_Files USE INDEX (DirID) WHERE DirID IN ( %s )" % dirListString
        result = self.db._query(req)
        if not result["OK"]:
            return result

        totalRecords = result["Value"][0][0]

        if not totalRecords:
            result = S_OK([])
            result["TotalRecords"] = totalRecords
            return result

        req = "SELECT FileID FROM FC_Files WHERE DirID IN ( %s ) LIMIT %s, %s " % (dirListString, startItem, maxItems)
        result = self.db._query(req)
        if not result["OK"]:
            return result
        result = S_OK([fileId[0] for fileId in result["Value"]])
        result["TotalRecords"] = totalRecords
        return result

    def getFileLFNsInDirectory(self, dirID, credDict):
        """Get file lfns for the given directory or directory list"""
        dirs = dirID
        if not isinstance(dirID, list):
            dirs = [dirID]

        dirListString = ",".join([str(dir) for dir in dirs])
        treeTable = self.getTreeTable()
        req = "SELECT CONCAT(D.DirName,'/',F.FileName) FROM FC_Files as F, %s as D WHERE D.DirID IN ( %s ) and D.DirID=F.DirID"
        req = req % (treeTable, dirListString)
        result = self.db._query(req)
        if not result["OK"]:
            return result
        lfnList = [x[0] for x in result["Value"]]
        return S_OK(lfnList)

    def getFileLFNsInDirectoryByDirectory(self, dirIDList, credDict):
        """Get file LFNs and IDs for the given directory or directory list

        :param list dirIDList: List of directory IDs
        :param dict credDict: dictionary of user credentials

        :return: S_OK/S_ERROR with Value dictionary {"DirLFNDict": dirLfnDict, "IDLFNDict": idLfnDict}
                 where dirLfnDict has the structure <directory_name>:<list of contained file names>,
                 idLfnDict has structure <fileID>:<LFN>
        """
        dirs = dirIDList
        if not isinstance(dirIDList, list):
            dirs = [dirIDList]

        dirListString = ",".join([str(dir_) for dir_ in dirs])
        treeTable = self.getTreeTable()
        req = "SELECT D.DirName,F.FileName,F.FileID FROM FC_Files as F, %s as D WHERE D.DirID IN ( %s ) and D.DirID=F.DirID"
        req = req % (treeTable, dirListString)
        result = self.db._query(req)
        if not result["OK"]:
            return result

        dirLfnDict = {}
        idLfnDict = {}
        for dir_, fname, fileID in result["Value"]:
            dirLfnDict.setdefault(dir_, []).append(fname)
            idLfnDict[fileID] = dir_ + "/" + fname

        return S_OK({"DirLFNDict": dirLfnDict, "IDLFNDict": idLfnDict})

    def _getDirectoryContents(self, path, details=False):
        """Get contents of a given directory"""
        result = self.findDir(path)
        if not result["OK"]:
            return result
        directoryID = result["Value"]
        directories = {}
        files = {}
        links = {}
        result = self.getChildren(path)
        if not result["OK"]:
            return result

        # Get subdirectories
        dirIDList = result["Value"]
        for dirID in dirIDList:
            result = self.getDirectoryPath(dirID)
            if not result["OK"]:
                return result
            dirName = result["Value"]
            if details:
                result = self.getDirectoryParameters(dirID)
                if not result["OK"]:
                    directories[dirName] = False
                else:
                    directories[dirName] = result["Value"]
            else:
                directories[dirName] = True
        result = self.db.fileManager.getFilesInDirectory(directoryID, verbose=details)
        if not result["OK"]:
            return result
        files = result["Value"]
        result = self.db.datasetManager.getDatasetsInDirectory(directoryID, verbose=details)
        if not result["OK"]:
            return result
        datasets = result["Value"]
        pathDict = {"Files": files, "SubDirs": directories, "Links": links, "Datasets": datasets}

        return S_OK(pathDict)

    def listDirectory(self, lfns, verbose=False):
        """Get the directory listing"""
        successful = {}
        failed = {}
        for path in lfns:
            result = self._getDirectoryContents(path, details=verbose)
            if not result["OK"]:
                failed[path] = result["Message"]
            else:
                successful[path] = result["Value"]

        return S_OK({"Successful": successful, "Failed": failed})

    def getDirectoryReplicas(self, lfns, allStatus=False):
        """Get replicas for files in the given directories"""
        successful = {}
        failed = {}
        for path in lfns:
            result = self.findDir(path)
            if not result["OK"]:
                failed[path] = result["Message"]
                continue
            directoryID = result["Value"]
            result = self.db.fileManager.getDirectoryReplicas(directoryID, path, allStatus)
            if not result["OK"]:
                failed[path] = result["Message"]
                continue
            fileDict = result["Value"]
            successful[path] = {}
            for fileName in fileDict:
                successful[path][fileName] = fileDict[fileName]

        return S_OK({"Successful": successful, "Failed": failed})

    def getDirectorySize(self, lfns, longOutput=False, rawFileTables=False, recursiveSum=True):
        """
        Get the total size of the requested directories. If longOutput flag
        is True, get also physical size per Storage Element

        :param bool longOutput: if True, also fetches the physical size per SE
        :param bool rawFileTables: if True, uses the File table instead of the pre-computed values
        :param bool recursiveSum: if True (default), takes into account subdirectories

        """
        start = time.time()

        result = self.db._getConnection()
        if not result["OK"]:
            return result
        connection = result["Value"]

        if rawFileTables:
            resultLogical = self._getDirectoryLogicalSize(lfns, recursiveSum=recursiveSum, connection=connection)
        else:
            resultLogical = self._getDirectoryLogicalSizeFromUsage(
                lfns, recursiveSum=recursiveSum, connection=connection
            )

        if not resultLogical["OK"]:
            connection.close()
            return resultLogical

        resultDict = resultLogical["Value"]
        if not resultDict["Successful"]:
            connection.close()
            return resultLogical

        if longOutput:
            # Continue with only successful directories
            if rawFileTables:
                resultPhysical = self._getDirectoryPhysicalSize(
                    resultDict["Successful"], recursiveSum=recursiveSum, connection=connection
                )
            else:
                resultPhysical = self._getDirectoryPhysicalSizeFromUsage(
                    resultDict["Successful"], recursiveSum=recursiveSum, connection=connection
                )
            if not resultPhysical["OK"]:
                resultDict["QueryTime"] = time.time() - start
                result = S_OK(resultDict)
                result["Message"] = "Failed to get the physical size on storage"
                connection.close()
                return result
            for lfn in resultPhysical["Value"]["Successful"]:
                resultDict["Successful"][lfn]["PhysicalSize"] = resultPhysical["Value"]["Successful"][lfn]
        connection.close()
        resultDict["QueryTime"] = time.time() - start

        return S_OK(resultDict)

    def _getDirectoryLogicalSizeFromUsage(self, lfns, recursiveSum=True, connection=None):
        """Get the total "logical" size of the requested directories

        :param recursiveSum: If false, don't take subdir into account
        """

        if not recursiveSum:
            return S_ERROR("Not implemented")

        successful = {}
        failed = {}
        for path in lfns:
            result = self.findDir(path)
            if not result["OK"]:
                failed[path] = "Directory not found"
                continue
            if not result["Value"]:
                failed[path] = "Directory not found"
                continue
            dirID = result["Value"]
            req = "SELECT SESize, SEFiles FROM FC_DirectoryUsage WHERE SEID=0 AND DirID=%d" % dirID

            result = self.db._query(req, connection)
            if not result["OK"]:
                failed[path] = result["Message"]
            elif not result["Value"]:
                successful[path] = {"LogicalSize": 0, "LogicalFiles": 0, "LogicalDirectories": 0}
            elif result["Value"][0][0]:
                successful[path] = {
                    "LogicalSize": int(result["Value"][0][0]),
                    "LogicalFiles": int(result["Value"][0][1]),
                }
                result = self.countSubdirectories(dirID, includeParent=False)
                if result["OK"]:
                    successful[path]["LogicalDirectories"] = result["Value"]
                else:
                    successful[path]["LogicalDirectories"] = -1

            else:
                successful[path] = {"LogicalSize": 0, "LogicalFiles": 0, "LogicalDirectories": 0}

        return S_OK({"Successful": successful, "Failed": failed})

    def _getDirectoryLogicalSize(self, lfns, recursiveSum=True, connection=None):
        """Get the total "logical" size of the requested directories

        :param bool recursiveSum: If false, don't take subdir into account
        """

        if not recursiveSum:
            return S_ERROR("Not implemented")

        successful = {}
        failed = {}
        treeTable = self.getTreeTable()
        for path in lfns:

            if path == "/":
                req = "SELECT SUM(Size),COUNT(*) FROM FC_Files"
                reqDir = "SELECT count(*) FROM %s" % treeTable
            else:
                result = self.findDir(path)
                if not result["OK"]:
                    failed[path] = "Directory not found"
                    continue
                if not result["Value"]:
                    failed[path] = "Directory not found"
                    continue
                dirID = result["Value"]
                result = self.getSubdirectoriesByID(dirID, requestString=True, includeParent=True)
                if not result["OK"]:
                    failed[path] = result["Message"]
                    continue
                else:
                    dirString = result["Value"]
                    req = (
                        "SELECT SUM(F.Size),COUNT(*) FROM FC_Files as F JOIN (%s) as T WHERE F.DirID=T.DirID"
                        % dirString
                    )
                    reqDir = dirString.replace("SELECT DirID FROM", "SELECT count(*) FROM")

            result = self.db._query(req, connection)
            if not result["OK"]:
                failed[path] = result["Message"]
            elif not result["Value"]:
                successful[path] = {"LogicalSize": 0, "LogicalFiles": 0, "LogicalDirectories": 0}
            elif result["Value"][0][0]:
                successful[path] = {
                    "LogicalSize": int(result["Value"][0][0]),
                    "LogicalFiles": int(result["Value"][0][1]),
                }
                result = self.db._query(reqDir, connection)
                if result["OK"] and result["Value"]:
                    successful[path]["LogicalDirectories"] = result["Value"][0][0] - 1
                else:
                    successful[path]["LogicalDirectories"] = -1

            else:
                successful[path] = {"LogicalSize": 0, "LogicalFiles": 0, "LogicalDirectories": 0}

        return S_OK({"Successful": successful, "Failed": failed})

    def _getDirectoryPhysicalSizeFromUsage(self, lfns, recursiveSum=True, connection=None):
        """Get the total size of the requested directories

        :param recursiveSum: If false, don't take subdir into account
        """

        if not recursiveSum:
            return S_ERROR("Not implemented")

        successful = {}
        failed = {}
        for path in lfns:
            result = self.findDir(path)
            if not result["OK"]:
                failed[path] = "Directory not found"
                continue
            if not result["Value"]:
                failed[path] = "Directory not found"
                continue
            dirID = result["Value"]

            req = "SELECT S.SEID, S.SEName, D.SESize, D.SEFiles FROM FC_DirectoryUsage as D, FC_StorageElements as S"
            req += "  WHERE S.SEID=D.SEID AND D.DirID=%d" % dirID
            result = self.db._query(req, connection)
            if not result["OK"]:
                failed[path] = result["Message"]
            elif not result["Value"]:
                successful[path] = {}
            elif result["Value"][0][0]:
                seDict = {}
                totalSize = 0
                totalFiles = 0
                for seID, seName, seSize, seFiles in result["Value"]:
                    if seSize or seFiles:
                        seDict[seName] = {"Size": seSize, "Files": seFiles}
                        totalSize += seSize
                        totalFiles += seFiles
                    else:
                        req = "DELETE FROM FC_DirectoryUsage WHERE SEID=%d AND DirID=%d" % (seID, dirID)
                        result = self.db._update(req)
                        if not result["OK"]:
                            gLogger.error("Failed to delete entry from FC_DirectoryUsage", result["Message"])
                seDict["TotalSize"] = int(totalSize)
                seDict["TotalFiles"] = int(totalFiles)
                successful[path] = seDict
            else:
                successful[path] = {}

        return S_OK({"Successful": successful, "Failed": failed})

    def _getDirectoryPhysicalSizeFromUsage_old(self, lfns, connection):
        """Get the total size of the requested directories"""
        successful = {}
        failed = {}
        for path in lfns:

            if path == "/":
                req = "SELECT S.SEName, D.SESize, D.SEFiles FROM FC_DirectoryUsage as D, FC_StorageElements as S"
                req += "  WHERE S.SEID=D.SEID"
            else:
                result = self.findDir(path)
                if not result["OK"]:
                    failed[path] = "Directory not found"
                    continue
                if not result["Value"]:
                    failed[path] = "Directory not found"
                    continue
                dirID = result["Value"]
                result = self.getSubdirectoriesByID(dirID, requestString=True, includeParent=True)
                if not result["OK"]:
                    return result
                subDirString = result["Value"]
                req = "SELECT S.SEName, D.SESize, D.SEFiles FROM FC_DirectoryUsage as D, FC_StorageElements as S"
                req += " JOIN (%s) AS F" % subDirString
                req += " WHERE S.SEID=D.SEID AND D.DirID=F.DirID"

            result = self.db._query(req, connection)
            if not result["OK"]:
                failed[path] = result["Message"]
            elif not result["Value"]:
                successful[path] = {}
            elif result["Value"][0][0]:
                seDict = {}
                totalSize = 0
                totalFiles = 0
                for seName, seSize, seFiles in result["Value"]:
                    sfDict = seDict.get(seName, {"Size": 0, "Files": 0})
                    sfDict["Size"] += seSize
                    sfDict["Files"] += seFiles
                    seDict[seName] = sfDict
                    totalSize += seSize
                    totalFiles += seFiles
                seDict["TotalSize"] = int(totalSize)
                seDict["TotalFiles"] = int(totalFiles)
                successful[path] = seDict
            else:
                successful[path] = {}

        return S_OK({"Successful": successful, "Failed": failed})

    def _getDirectoryPhysicalSize(self, lfns, recursiveSum=True, connection=None):
        """Get the total size of the requested directories
        :param recursiveSum: If false, don't take subdir into account
        """
        if not recursiveSum:
            return S_ERROR("Not implemented")

        successful = {}
        failed = {}
        for path in lfns:
            if path == "/":
                req = "SELECT SUM(F.Size),COUNT(F.Size),S.SEName from FC_Files as F, FC_Replicas as R, FC_StorageElements as S "
                req += "WHERE R.SEID=S.SEID AND F.FileID=R.FileID "
                req += "GROUP BY S.SEID"
            else:
                result = self.findDir(path)
                if not result["OK"]:
                    failed[path] = "Directory not found"
                    continue
                if not result["Value"]:
                    failed[path] = "Directory not found"
                    continue
                dirID = result["Value"]
                result = self.getSubdirectoriesByID(dirID, requestString=True, includeParent=True)
                if not result["OK"]:
                    failed[path] = result["Message"]
                    continue
                else:
                    dirString = result["Value"]

                    req = (
                        "SELECT SUM(F.Size),COUNT(F.Size),S.SEName from FC_Files as F, FC_Replicas as R, FC_StorageElements as S JOIN (%s) as T "
                        % dirString
                    )
                    req += "WHERE R.SEID=S.SEID AND F.FileID=R.FileID AND F.DirID=T.DirID "
                    req += "GROUP BY S.SEID"

            result = self.db._query(req, connection)
            if not result["OK"]:
                failed[path] = result["Message"]
            elif not result["Value"]:
                successful[path] = {}
            elif result["Value"][0][0]:
                seDict = {}
                totalSize = 0
                totalFiles = 0
                for size, files, seName in result["Value"]:
                    seDict[seName] = {"Size": int(size), "Files": int(files)}
                    totalSize += size
                    totalFiles += files
                seDict["TotalSize"] = int(totalSize)
                seDict["TotalFiles"] = int(totalFiles)
                successful[path] = seDict
            else:
                successful[path] = {}

        return S_OK({"Successful": successful, "Failed": failed})

    def _rebuildDirectoryUsage(self):
        """Recreate and replenish the Storage Usage tables"""

        req = "DROP TABLE IF EXISTS FC_DirectoryUsage_backup"
        result = self.db._update(req)
        req = "RENAME TABLE FC_DirectoryUsage TO FC_DirectoryUsage_backup"
        result = self.db._update(req)
        req = "CREATE TABLE `FC_DirectoryUsage` LIKE `FC_DirectoryUsage_backup`"
        result = self.db._update(req)
        if not result["OK"]:
            return result

        result = self.__rebuildDirectoryUsageLeaves()
        if not result["OK"]:
            return result

        result = self.db.dtree.findDir("/")
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directory / not found")
        dirID = result["Value"]
        result = self.__rebuildDirectoryUsage(dirID)
        gLogger.verbose("Finished rebuilding Directory Usage")
        return result

    def __rebuildDirectoryUsageLeaves(self):
        """Rebuild DirectoryUsage entries for directories having files"""
        req = "SELECT DISTINCT(DirID) FROM FC_Files"
        result = self.db._query(req)
        if not result["OK"]:
            return result

        dirIDs = [x[0] for x in result["Value"]]
        gLogger.verbose("Starting rebuilding Directory Usage, number of visible directories %d" % len(dirIDs))

        insertFields = ["DirID", "SEID", "SESize", "SEFiles", "LastUpdate"]
        insertCount = 0
        insertValues = []

        count = 0
        empty = 0

        for dirID in dirIDs:

            count += 1

            # Get the physical size
            req = "SELECT SUM(F.Size),COUNT(F.Size),R.SEID from FC_Files as F, FC_Replicas as R "
            req += "WHERE F.FileID=R.FileID AND F.DirID=%d GROUP BY R.SEID" % int(dirID)
            result = self.db._query(req)
            if not result["OK"]:
                return result
            if not result["Value"]:
                empty += 1

            for seSize, seFiles, seID in result["Value"]:
                insertValues = [dirID, seID, seSize, seFiles, "UTC_TIMESTAMP()"]
                result = self.db.insertFields("FC_DirectoryUsage", insertFields, insertValues)
                if not result["OK"]:
                    if "Duplicate" in result["Message"]:
                        req = "UPDATE FC_DirectoryUsage SET SESize=%d, SEFiles=%d, LastUpdate=UTC_TIMESTAMP()" % (
                            seSize,
                            seFiles,
                        )
                        req += " WHERE DirID=%s AND SEID=%s" % (dirID, seID)
                        result = self.db._update(req)
                        if not result["OK"]:
                            return result
                    return result

            # Get the logical size
            req = "SELECT SUM(Size),COUNT(Size) from FC_Files WHERE DirID=%d " % int(dirID)
            result = self.db._query(req)
            if not result["OK"]:
                return result
            if not result["Value"]:
                return S_ERROR("Empty directory")
            seSize, seFiles = result["Value"][0]
            insertValues = [dirID, 0, seSize, seFiles, "UTC_TIMESTAMP()"]
            result = self.db.insertFields("FC_DirectoryUsage", insertFields, insertValues)
            if not result["OK"]:
                if "Duplicate" in result["Message"]:
                    req = "UPDATE FC_DirectoryUsage SET SESize=%d, SEFiles=%d, LastUpdate=UTC_TIMESTAMP()" % (
                        seSize,
                        seFiles,
                    )
                    req += " WHERE DirID=%s AND SEID=0" % dirID
                    result = self.db._update(req)
                    if not result["OK"]:
                        return result
                else:
                    return result

        gLogger.verbose("Processed %d directories, %d empty " % (count, empty))

        return S_OK()

    def __rebuildDirectoryUsage(self, directoryID):
        """Rebuild DirectoryUsage entries recursively for the given path"""
        result = self.getChildren(directoryID)
        if not result["OK"]:
            return result
        dirIDs = result["Value"]
        resultDict = {}
        for dirID in dirIDs:
            result = self.__rebuildDirectoryUsage(dirID)
            if not result["OK"]:
                return result
            dirDict = result["Value"]
            for seID in dirDict:
                resultDict.setdefault(seID, {"Size": 0, "Files": 0})
                resultDict[seID]["Size"] += dirDict[seID]["Size"]
                resultDict[seID]["Files"] += dirDict[seID]["Files"]

        insertFields = ["DirID", "SEID", "SESize", "SEFiles", "LastUpdate"]
        insertValues = []
        for seID in resultDict:
            size = resultDict[seID]["Size"]
            files = resultDict[seID]["Files"]
            req = "UPDATE FC_DirectoryUsage SET SESize=SESize+%d, SEFiles=SEFiles+%d WHERE DirID=%d AND SEID=%d"
            req = req % (size, files, directoryID, seID)
            result = self.db._update(req)
            if not result["OK"]:
                return result
            if not result["Value"]:
                insertValues = [directoryID, seID, size, files, "UTC_TIMESTAMP()"]
                result = self.db.insertFields("FC_DirectoryUsage", insertFields, insertValues)
                if not result["OK"]:
                    return result

        req = "SELECT SEID,SESize,SEFiles from FC_DirectoryUsage WHERE DirID=%d" % directoryID
        result = self.db._query(req)
        if not result["OK"]:
            return result

        resultDict = {}
        for seid, size, files in result["Value"]:
            resultDict[seid] = {"Size": size, "Files": files}

        return S_OK(resultDict)

    def getDirectoryCounters(self, connection=False):
        """Get the total number of directories"""
        conn = self._getConnection(connection)
        resultDict = {}
        req = "SELECT COUNT(*) from FC_DirectoryInfo"
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        resultDict["Directories"] = res["Value"][0][0]

        treeTable = self.getTreeTable()

        req = "SELECT COUNT(DirID) FROM %s WHERE Parent NOT IN ( SELECT DirID from %s )" % (treeTable, treeTable)
        req += " AND DirID <> 1"
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        resultDict["Orphan Directories"] = res["Value"][0][0]

        req = "SELECT COUNT(DirID) FROM %s WHERE DirID NOT IN ( SELECT Parent from %s )" % (treeTable, treeTable)
        req += " AND DirID NOT IN ( SELECT DirID from FC_Files ) "
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        resultDict["Empty Directories"] = res["Value"][0][0]

        req = "SELECT COUNT(DirID) FROM %s WHERE DirID NOT IN ( SELECT DirID FROM FC_DirectoryInfo )" % treeTable
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        resultDict["DirTree w/o DirInfo"] = res["Value"][0][0]

        req = "SELECT COUNT(DirID) FROM FC_DirectoryInfo WHERE DirID NOT IN ( SELECT DirID FROM %s )" % treeTable
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        resultDict["DirInfo w/o DirTree"] = res["Value"][0][0]

        return S_OK(resultDict)



""" Refresh local CS (if needed)
Used each time you call gConfig. It keep your configuration up-to-date with the configuration server
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import threading
import time
import os

from six.moves import _thread as thread

from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.private.RefresherBase import RefresherBase
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.Core.Utilities import LockRing


class Refresher(RefresherBase, threading.Thread):
    """
    The refresher
    A long time ago, in a code away, far away...
    A guy do the code to autorefresh the configuration
    To prepare transition to HTTPS we have done separation
    between the logic and the implementation of background
    tasks, it's the original version, for diset, using thread.

    """

    def __init__(self):
        threading.Thread.__init__(self)
        RefresherBase.__init__(self)
        self._triggeredRefreshLock = LockRing.LockRing().getLock()

    def _refreshInThread(self):
        """
        Refreshing configuration in the background. By default it uses a thread but it can be
        run also in the IOLoop
        """
        retVal = self._refresh()
        if not retVal["OK"]:
            gLogger.error("Error while updating the configuration", retVal["Message"])

    def refreshConfigurationIfNeeded(self):
        """
        Refresh the configuration if automatic updates are disabled, refresher is enabled and servers are defined
        """
        if not self._refreshEnabled or self._automaticUpdate or not gConfigurationData.getServers():
            return
        # To improve performance, skip acquiring the lock if possible
        if not self._lastRefreshExpired():
            return
        self._triggeredRefreshLock.acquire()
        try:
            if not self._lastRefreshExpired():
                return
            self._lastUpdateTime = time.time()
        finally:
            try:
                self._triggeredRefreshLock.release()
            except thread.error:
                pass
        # Launch the refresh
        thd = threading.Thread(target=self._refreshInThread)
        thd.setDaemon(1)
        thd.start()

    def autoRefreshAndPublish(self, sURL):
        """
        Start the autorefresh background task

        :param str sURL: URL of the configuration server
        """
        gLogger.debug("Setting configuration refresh as automatic")
        if not gConfigurationData.getAutoPublish():
            gLogger.debug("Slave server won't auto publish itself")
        if not gConfigurationData.getName():
            import DIRAC

            DIRAC.abort(10, "Missing configuration name!")
        self._url = sURL
        self._automaticUpdate = True
        self.setDaemon(1)
        self.start()

    def run(self):
        while self._automaticUpdate:
            time.sleep(gConfigurationData.getPropagationTime())
            if self._refreshEnabled:
                if not self._refreshAndPublish():
                    gLogger.error("Can't refresh configuration from any source")

    def daemonize(self):
        """
        Daemonize the background tasks
        """

        self.setDaemon(1)
        self.start()


# Here we define the refresher which should be used.
# By default we use the original refresher.

# Be careful, if you never start the IOLoop (with a TornadoServer for example)
# the TornadoRefresher will not work. IOLoop can be started after refresher
# but background tasks will be delayed until IOLoop start.
# DIRAC_USE_TORNADO_IOLOOP is defined by starting scripts
if os.environ.get("DIRAC_USE_TORNADO_IOLOOP", "false").lower() in ("yes", "true"):
    from DIRAC.ConfigurationSystem.private.TornadoRefresher import TornadoRefresher

    gRefresher = TornadoRefresher()
else:
    gRefresher = Refresher()


if __name__ == "__main__":
    time.sleep(0.1)
    gRefresher.daemonize()



""" Exposes singleton object gMonitor, which is instance of MonitoringClient class

    Uses RPC Framework/Monitoring service. Calls registerActivities exposed function
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id"

import time
import os
import six

import DIRAC
from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.DEncode import ignoreEncodeWarning
from DIRAC.Core.Utilities.LockRing import LockRing
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.Utilities import Time, Network, ThreadScheduler
from DIRAC.Core.DISET.RPCClient import RPCClient


class MonitoringClientActivityNotDefined(Exception):
    """This class is used to raise an exception if an activity is not defined meaning
    not registered using the registerActivity method of the gMonitor object.
    """

    def __init__(self, message):
        self.message = str(message)

    def __str__(self):
        return self.message


class MonitoringClientActivityValueTypeError(Exception):
    """This class is used to raise an exception if an activity type is mismatched while
    calling the addMark method of the gMonitor object.
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class MonitoringClientUnknownParameter(Exception):
    """This class is used to raise an exception when some unkown parameter is passed
    to the setComponentExtraParam method of the gMonitor object.
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class MonitoringFlusher(object):
    """
    This class is used to flush all the instances of the MonitoringClient registered periodically.
    """

    def __init__(self):
        self.__mcList = []
        ThreadScheduler.gThreadScheduler.addPeriodicTask(300, self.flush)
        # HACK: Avoid exiting while the thread is starting
        time.sleep(0.1)

    def flush(self, allData=False):
        """This method is used to periodically flush the data.

        :type allData: bool
        :param allData: This is used to indicate whether all the data is present or not.
        """
        for mc in self.__mcList:
            mc.flush(allData)

    def registerMonitoringClient(self, mc):
        """This method is used to register instances of the MonitoringClient.

        :type mc: object of MonitoringClient
        :param mc: Its just an instance of the MonitoringClient that we want to register.
        """
        if mc not in self.__mcList:
            self.__mcList.append(mc)


# DIRAC_USE_TORNADO_IOLOOP is defined by starting scripts
if os.environ.get("DIRAC_USE_TORNADO_IOLOOP", "false").lower() in ("yes", "true"):
    from DIRAC.FrameworkSystem.Client.MonitoringClientIOLoop import MonitoringFlusherTornado

    gMonitoringFlusher = MonitoringFlusherTornado()
else:
    gMonitoringFlusher = MonitoringFlusher()


class MonitoringClient(object):
    """This class is used to create the gMonitor object and acts as a client side for registering activities
    and committing the data.
    """

    # Different types of operations
    OP_MEAN = "mean"
    OP_ACUM = "acum"
    OP_SUM = "sum"
    OP_RATE = "rate"

    # Predefined components that can be registered
    COMPONENT_SERVICE = "service"
    COMPONENT_AGENT = "agent"
    COMPONENT_WEB = "web"
    COMPONENT_SCRIPT = "script"
    COMPONENT_TORNADO = "tornado"

    __validMonitoringValues = six.integer_types + (float,)

    def __init__(self):
        self.sourceId = 0
        self.sourceDict = {}
        self.sourceDict["componentType"] = "unknown"
        self.sourceDict["componentName"] = "unknown"
        self.sourceDict["componentLocation"] = "unknown"
        self.activitiesDefinitions = {}
        self.activitiesMarks = {}
        self.definitionsToSend = {}
        self.marksToSend = {}
        self.__compRegistrationExtraDict = {}
        self.__compCommitExtraDict = {}
        self.__activitiesLock = None  # threading.Lock()
        self.__flushingLock = None  # threading.Lock()
        self.timeStep = 60
        self.__initialized = False
        self.__enabled = True

    @property
    def activitiesLock(self):
        if not self.__activitiesLock:
            self.__activitiesLock = LockRing().getLock("activityLock")
        return self.__activitiesLock

    @property
    def flushingLock(self):
        if not self.__flushingLock:
            self.__flushingLock = LockRing().getLock("flushingLock")
        return self.__flushingLock

    def disable(self):
        self.__enabled = False

    def enable(self):
        self.__enabled = True

    def setComponentExtraParam(self, name, value):
        """
        Sets the extra parameters of the component reporting.

        :type  name: string
        :param name: It should belong to one of these 'version', 'DIRACVersion', 'description', 'startTime', 'platform',
                                                      'cycles', 'queries'.
        :type value: string
        :param value: The proper value corresponding to one of the selected names.
        """
        if name in ("version", "DIRACVersion", "description", "startTime", "platform"):
            self.__compRegistrationExtraDict[name] = str(value)
        elif name in ("cycles", "queries", "connections"):
            self.__compCommitExtraDict[name] = str(value)
        else:
            raise MonitoringClientUnknownParameter("Unknown parameter %s" % name)
            # raise Exception( "Unknown parameter %s" % name )

    def initialize(self):
        self.logger = gLogger.getSubLogger("Monitoring")
        self.logger.debug("Initializing Monitoring Client")
        self.sourceDict["setup"] = gConfig.getValue("/DIRAC/Setup")
        self.sourceDict["site"] = DIRAC.siteName()
        if self.sourceDict["componentType"] == self.COMPONENT_SERVICE:
            self.cfgSection = PathFinder.getSystemSection(self.sourceDict["componentName"])
        elif self.sourceDict["componentType"] == self.COMPONENT_AGENT:
            self.cfgSection = PathFinder.getAgentSection(self.sourceDict["componentName"])
            self.setComponentLocation(Network.getFQDN())
        elif self.sourceDict["componentType"] == self.COMPONENT_WEB:
            self.cfgSection = "/WebApp"
            self.setComponentLocation("http://%s" % Network.getFQDN())
            self.setComponentName("WebApp")
        elif self.sourceDict["componentType"] == self.COMPONENT_SCRIPT:
            self.cfgSection = "/Script"
        elif self.sourceDict["componentType"] == self.COMPONENT_TORNADO:
            self.cfgSection = "/Tornado"
        else:
            raise Exception("Component type has not been defined")
        gMonitoringFlusher.registerMonitoringClient(self)
        self.__initialized = True

    def setComponentLocation(self, componentLocation=False):
        """
        Sets the location of the component reporting.

        :type  componentLocation: string
        :param componentLocation: Location of the component reporting
        """
        if not componentLocation:
            self.sourceDict["componentLocation"] = gConfig.getValue("/Site")
        else:
            self.sourceDict["componentLocation"] = componentLocation

    def setComponentName(self, componentName):
        """
        Sets the name of the component reporting.

        :type  componentName: string
        :param componentName: Name of the component reporting
        """
        self.sourceDict["componentName"] = componentName

    def setComponentType(self, componentType):
        """
        Defines the type of component reporting data.

        :type  componentType: string
        :param componentType: Defines the grouping of the host by type. All the possibilities
                              are defined in the Constants.py file
        """
        self.sourceDict["componentType"] = componentType

    def registerActivity(self, name, description, category, unit, operation, bucketLength=60):
        """
        Registers new activity. Before reporting information to the server, the activity
        must be registered.

        :type  name: string
        :param name: Id of the activity to report
        :type description: string
        :param description: Description of the activity
        :type  category: string
        :param category: Grouping of the activity
        :type  unit: string
        :param unit: String representing the unit that will be printed in the plots
        :type  operation: string
        :param operation: Type of data operation to represent data. All the possibilities
                          are defined in the Constants.py file
        :type  bucketLength: int
        :param bucketLength: Bucket length in seconds
        """
        if not self.__initialized:
            return
        self.activitiesLock.acquire()
        try:
            self.logger.debug("Registering activity %s" % name)
            if name not in self.activitiesDefinitions:
                self.activitiesDefinitions[name] = {
                    "category": category,
                    "description": description,
                    "unit": unit,
                    "type": operation,
                    "bucketLength": bucketLength,
                }
                self.activitiesMarks[name] = {}
                self.definitionsToSend[name] = dict(self.activitiesDefinitions[name])
        finally:
            self.activitiesLock.release()

    def __UTCStepTime(self, acName):
        stepLength = self.activitiesDefinitions[acName]["bucketLength"]
        nowEpoch = int(Time.toEpoch())
        return nowEpoch - nowEpoch % stepLength

    def addMark(self, name, value=1):
        """
        Adds a new mark to the specified activity

        :type  name: string
        :param name: Name of the activity to report
        :type  value: number
        :param value: Weight of the mark. By default it's one.
        """
        if not self.__initialized:
            return
        if not self.__enabled:
            return
        if name not in self.activitiesDefinitions:
            raise MonitoringClientActivityNotDefined("You must register activity %s before adding marks to it" % name)
            # raise Exception( "You must register activity %s before adding marks to it" % name )
        if not isinstance(value, self.__validMonitoringValues):
            raise MonitoringClientActivityValueTypeError(
                "Activity '%s' value's type (%s) is not valid" % (name, type(value))
            )
            # raise Exception( "Value's type %s is not valid" % value )
        self.activitiesLock.acquire()
        try:
            self.logger.debug("Adding mark to %s" % name)
            markTime = self.__UTCStepTime(name)
            if markTime in self.activitiesMarks[name]:
                self.activitiesMarks[name][markTime].append(value)
            else:
                self.activitiesMarks[name][markTime] = [value]
        finally:
            self.activitiesLock.release()

    def __consolidateMarks(self, allData):
        """
        Copies all marks except last step ones and consolidates them.

        :type allData: bool
        :param allData: This is used to indicate whether all the data is present or not.
        :return: dictionary of consolidatedMarks.
        """
        consolidatedMarks = {}
        remainderMarks = {}
        for key in self.activitiesMarks:
            if allData:
                lastStepToSend = int(Time.toEpoch())
            else:
                lastStepToSend = self.__UTCStepTime(key)
            consolidatedMarks[key] = {}
            remainderMarks[key] = {}
            for markTime in self.activitiesMarks[key]:
                markValue = self.activitiesMarks[key][markTime]
                if markTime >= lastStepToSend:
                    remainderMarks[key][markTime] = markValue
                else:
                    consolidatedMarks[key][markTime] = markValue
                    # Consolidate the copied ones
                    totalValue = 0
                    for mark in consolidatedMarks[key][markTime]:
                        totalValue += mark
                    if self.activitiesDefinitions[key]["type"] == self.OP_MEAN:
                        totalValue /= len(consolidatedMarks[key][markTime])
                    consolidatedMarks[key][markTime] = totalValue
            if len(consolidatedMarks[key]) == 0:
                del consolidatedMarks[key]
        self.activitiesMarks = remainderMarks
        return consolidatedMarks

    def flush(self, allData=False):
        """This method is used to periodically flush the data and send it to the server side.

        :type allData: bool
        :param allData: This is used to indicate whether all the data is present or not.
        """
        if not self.__enabled or not self.__initialized:
            return
        self.flushingLock.acquire()
        self.logger.debug("Sending information to server")
        try:
            self.activitiesLock.acquire()
            try:
                self.logger.debug("Consolidating data...")
                self.__appendMarksToSend(self.__consolidateMarks(allData))
            finally:
                self.activitiesLock.release()
            # Commit new activities
            if self.__dataToSend():
                if not self.__disabled():
                    # this also creates a point of contact between te client and the server as it calls the __sendData method.
                    self.__sendData()
            self.__pruneMarksData()
        finally:
            self.flushingLock.release()

    def __disabled(self):
        """This method is basically used to check whether monitoring is disabled or not inside the
        configuration file.
        """
        return gConfig.getValue("%s/DisableMonitoring" % self.cfgSection, "false").lower() in ("yes", "y", "true", "1")

    def __dataToSend(self):
        return len(self.definitionsToSend) or len(self.marksToSend)

    def __appendMarksToSend(self, acMarks):
        if self.__disabled():
            return
        for acName in acMarks:
            if acName in self.marksToSend:
                for timeMark in acMarks[acName]:
                    self.marksToSend[acName][timeMark] = acMarks[acName][timeMark]
            else:
                self.marksToSend[acName] = acMarks[acName]

    # Decoding was fixed in https://github.com/DIRACGrid/DIRAC/pull/4462
    @ignoreEncodeWarning
    def __sendData(self, secsTimeout=False):
        """This method is used to initialize the rpcClient from the server and also to initiate the task
        of registering the activities and committing the marks.

        :type secsTimeout: int
        :param secsTimeout: The timeout in seconds for the rpcClient.
        """
        from DIRAC.FrameworkSystem.private.monitoring.ServiceInterface import gServiceInterface

        if gServiceInterface.srvUp:
            self.logger.debug("Using internal interface to send data")
            rpcClient = gServiceInterface
        else:
            self.logger.debug("Creating RPC client")
            # Here is where the client is created from the running Framework/Monitoring service.
            rpcClient = RPCClient("Framework/Monitoring", timeout=secsTimeout)
        # Send registrations
        if not self.__sendRegistration(rpcClient):
            return False
        # Send marks
        maxIteration = 5
        if self.__sendMarks(rpcClient) and maxIteration:
            maxIteration -= 1
            if not self.__sendRegistration(rpcClient):
                return False

    def __pruneMarksData(self):
        """This basically prunes the marks which exceed the bucket length."""
        for acName in self.marksToSend:
            maxBuckets = int(86400 / self.activitiesDefinitions[acName]["bucketLength"])
            if len(self.marksToSend[acName]) > maxBuckets:
                timeSlots = sorted(self.marksToSend[acName])
                while len(self.marksToSend[acName]) > maxBuckets:
                    del self.marksToSend[acName][timeSlots.pop(0)]

    def __sendRegistration(self, rpcClient):
        """This method registers all the activities using the rpcClient.

        :type rpcClient: object of RPCClient
        :param rpcClient: This is used to access the methods within the service and register the activities.
        :return: bool
        """
        if not len(self.definitionsToSend):
            return True
        self.logger.debug("Registering activities")
        retDict = rpcClient.registerActivities(
            self.sourceDict, self.definitionsToSend, self.__compRegistrationExtraDict
        )
        if not retDict["OK"]:
            self.logger.error("Can't register activities", retDict["Message"])
            return False
        self.sourceId = retDict["Value"]
        self.definitionsToSend = {}
        return True

    def __sendMarks(self, rpcClient):
        """
        This method sends all the marks accumulated to the server using the rpcClient.

        :type rpcClient: object of RPCClient
        :param rpcClient: This is used to access the methods within the service and register the activities.
        :return: bool
        """
        assert self.sourceId
        self.logger.debug("Sending marks")
        retDict = rpcClient.commitMarks(self.sourceId, self.marksToSend, self.__compCommitExtraDict)
        if not retDict["OK"]:
            self.logger.error("Can't send activities marks", retDict["Message"])
            return False
        acMissedMarks = {}
        if len(retDict["Value"]) > 0:
            self.logger.debug("There are activities unregistered")
            for acName in retDict["Value"]:
                if acName in self.activitiesDefinitions:
                    self.definitionsToSend[acName] = dict(self.activitiesDefinitions[acName])
                    acMissedMarks[acName] = self.marksToSend[acName]
                else:
                    self.logger.debug("Server reported unregistered activity that does not exist")
        self.marksToSend = acMissedMarks
        return len(self.definitionsToSend)

    def forceFlush(self, exitCode):
        """This method can be used to force flush all the data directly without flushing the
        data periodically.
        """
        self.sendingMode = "none"
        self.flush(allData=True)

    def getComponentsStatus(self, condDict):
        """This method is used to get the status of the components.

        :type condDict: dictionary
        :param condDict: A condition dictionary.
        :return: S_OK with status and message about the component.
        """
        rpcClient = RPCClient("Framework/Monitoring", timeout=100)
        return rpcClient.getComponentsStatus(condDict)

    def __filterComponent(self, component, condDict):
        for key in condDict:
            if key not in component:
                return False
            condVal = condDict[key]
            componentVal = component[key]
            if isinstance(condVal, (list, tuple)):
                if componentVal not in condVal:
                    return False
            else:
                if componentVal != condVal:
                    return False
        return True

    def getComponentsStatusWebFormatted(self, condDict={}, sortingList=[], startItem=0, maxItems=0):
        result = self.getComponentsStatus(condDict)
        if not result["OK"]:
            return result
        compDict, fields = result["Value"]
        tabledData = []
        for setup in compDict:
            for cType in compDict[setup]:
                for name in compDict[setup][cType]:
                    for component in compDict[setup][cType][name]:
                        # How here we are. Now we need to filter the components
                        if not self.__filterComponent(component, condDict):
                            continue
                        # Add to tabledData!
                        row = []
                        for field in fields:
                            if field not in component:
                                row.append("")
                            else:
                                row.append(component[field])
                        tabledData.append(row)
        # We've got the data in table form
        # Now it's time to sort it
        if sortingList:
            sortingData = []
            sortField = sortingList[0][0]
            if sortField not in fields:
                return S_ERROR("Sorting field %s does not exist" % sortField)
            sortDirection = sortingList[0][1]
            fieldIndex = 0
            for i in range(len(fields)):
                if fields[i] == sortField:
                    fieldIndex = i
                    break
            for row in tabledData:
                sortingData.append((row[fieldIndex], row))
            sortingData.sort(key=lambda x: x[0])
            if sortDirection == "DESC":
                sortingData.reverse()
            tabledData = [row[1] for row in sortingData]
        # Now need to limit
        numRows = len(tabledData)
        tabledData = tabledData[startItem:]
        if maxItems:
            tabledData = tabledData[:maxItems]
        returnData = {
            "ParameterNames": fields,
            "Records": tabledData,
            "TotalRecords": numRows,
        }
        return S_OK(returnData)


# Here the singleton gMonitor object is created which is used in all other use cases.
gMonitor = MonitoringClient()



""" Recipes on how to create job plots
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.AccountingSystem.Client.Types.Job import Job
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter


class JobPlotter(BaseReporter):

    _typeName = "Job"
    _typeKeyFields = [dF[0] for dF in Job().definitionKeyFields]

    def _translateGrouping(self, grouping):
        if grouping == "Country":
            sqlRepr = 'upper( substring( %s, locate( ".", %s, length( %s ) - 4 ) + 1 ) )'
            return (sqlRepr, ["Site", "Site", "Site"], sqlRepr)
        elif grouping == "Grid":
            return ('substring_index( %s, ".", 1 )', ["Site"])
        else:
            return ("%s", [grouping])

    _reportCPUEfficiencyName = "CPU efficiency"

    def _reportCPUEfficiency(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s), SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "CPUTime", "ExecTime"],
        )

        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {
                "checkNone": True,
                "convertToGranularity": "sum",
                "calculateProportionalGauges": False,
                "consolidationFunction": self._efficiencyConsolidation,
            },
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        if len(dataDict) > 1:
            # Get the total for the plot
            selectFields = ("'Total', %s, %s, SUM(%s),SUM(%s)", ["startTime", "bucketLength", "CPUTime", "ExecTime"])

            retVal = self._getTimedData(
                reportRequest["startTime"],
                reportRequest["endTime"],
                selectFields,
                reportRequest["condDict"],
                reportRequest["groupingFields"],
                {
                    "scheckNone": True,
                    "convertToGranularity": "sum",
                    "calculateProportionalGauges": False,
                    "consolidationFunction": self._efficiencyConsolidation,
                },
            )
            if not retVal["OK"]:
                return retVal
            totalDict = retVal["Value"][0]
            self.stripDataField(totalDict, 0)
            for key in totalDict:
                dataDict[key] = totalDict[key]
        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotCPUEfficiency(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Job CPU efficiency by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
        }
        return self._generateQualityPlot(filename, plotInfo["data"], metadata)

    _reportCPUUsedName = "Cumulative CPU time"

    def _reportCPUUsed(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "CPUTime"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        dataDict = self._accumulate(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "time"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotCPUUsed(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "CPU used by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
            "sort_labels": "max_value",
        }
        return self._generateCumulativePlot(filename, plotInfo["graphDataDict"], metadata)

    _reportCPUUsageName = "CPU time"

    def _reportCPUUsage(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "CPUTime"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, _maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "time"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotCPUUsage(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "CPU usage by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateStackedLinePlot(filename, plotInfo["graphDataDict"], metadata)

    _reportNormCPUUsedName = "Cumulative Normalized CPU"

    def _reportNormCPUUsed(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "NormCPUTime"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        dataDict = self._accumulate(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "cpupower"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotNormCPUUsed(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Normalized CPU used by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
            "sort_labels": "max_value",
        }
        return self._generateCumulativePlot(filename, plotInfo["graphDataDict"], metadata)

    _reportNormCPUUsageName = "Normalized CPU power"

    def _reportNormCPUUsage(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "NormCPUTime"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, _maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "cpupower"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotNormCPUUsage(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Normalized CPU usage by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateStackedLinePlot(filename, plotInfo["graphDataDict"], metadata)

    _reportWallTimeName = "Wall time"

    def _reportWallTime(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "ExecTime"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, _maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "time"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotWallTime(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Wall Time by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateStackedLinePlot(filename, plotInfo["graphDataDict"], metadata)

    _reportRunningJobsName = "Running jobs"

    def _reportRunningJobs(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "ExecTime"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, _maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "jobs"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotRunningJobs(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Running jobs by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateStackedLinePlot(filename, plotInfo["graphDataDict"], metadata)

    _reportTotalCPUUsedName = "Pie plot of CPU used"

    def _reportTotalCPUUsed(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", SUM(%s)/86400",
            reportRequest["groupingFields"][1] + ["CPUTime"],
        )
        retVal = self._getSummaryData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]
        return S_OK({"data": dataDict})

    def _plotTotalCPUUsed(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "CPU days used by %s" % reportRequest["grouping"],
            "ylabel": "CPU days",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
        }
        return self._generatePiePlot(filename, plotInfo["data"], metadata)

    _reportAccumulatedWallTimeName = "Cumulative wall time"

    def _reportAccumulatedWallTime(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "ExecTime"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        dataDict = self._accumulate(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "time"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotAccumulatedWallTime(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Cumulative wall time by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
            "sort_labels": "max_value",
        }
        return self._generateCumulativePlot(filename, plotInfo["graphDataDict"], metadata)

    _reportTotalWallTimeName = "Pie plot of wall time usage"

    def _reportTotalWallTime(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", SUM(%s)/86400",
            reportRequest["groupingFields"][1] + ["ExecTime"],
        )
        retVal = self._getSummaryData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]
        return S_OK({"data": dataDict})

    def _plotTotalWallTime(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Wall time days used by %s" % reportRequest["grouping"],
            "ylabel": "CPU days",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
        }
        return self._generatePiePlot(filename, plotInfo["data"], metadata)

    ##
    #  Jobs
    ##

    _reportCumulativeNumberOfJobsName = "Cumulative executed jobs"

    def _reportCumulativeNumberOfJobs(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        dataDict = self._accumulate(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "jobs"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotCumulativeNumberOfJobs(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Cumulative Jobs by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
            "sort_labels": "max_value",
        }
        return self._generateCumulativePlot(filename, plotInfo["graphDataDict"], metadata)

    _reportNumberOfJobsName = "Job execution rate"

    def _reportNumberOfJobs(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, _maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "jobs"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotNumberOfJobs(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Jobs by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateStackedLinePlot(filename, plotInfo["graphDataDict"], metadata)

    _reportTotalNumberOfJobsName = "Pie plot of executed jobs"

    def _reportTotalNumberOfJobs(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", SUM(%s)",
            reportRequest["groupingFields"][1] + ["entriesInBucket"],
        )
        retVal = self._getSummaryData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]
        return S_OK({"data": dataDict})

    def _plotTotalNumberOfJobs(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Total Number of Jobs by %s" % reportRequest["grouping"],
            "ylabel": "Jobs",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
        }
        return self._generatePiePlot(filename, plotInfo["data"], metadata)

    ##
    # Proc bw
    ##

    _reportProcessingBandwidthName = "Processing bandwidth"

    def _reportProcessingBandwidth(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM((%s)/(%s))/SUM(%s)",
            reportRequest["groupingFields"][1]
            + ["startTime", "bucketLength", "InputDataSize", "CPUTime", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, _maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "bytes"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotProcessingBandwidth(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Processing Bandwidth by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateStackedLinePlot(filename, plotInfo["graphDataDict"], metadata)

    ##
    # Data sizes
    ##

    _reportInputSandboxSizeName = "Input sandbox"

    def _reportInputSandboxSize(self, reportRequest):
        return self.__reportFieldSizeinMB(reportRequest, ("InputSandBoxSize", "Input sand box size"))

    _reportOutputSandboxSizeName = "Ouput sandbox"

    def _reportOutputSandboxSize(self, reportRequest):
        return self.__reportFieldSizeinMB(reportRequest, ("OutputSandBoxSize", "Output sand box size"))

    _reportDiskSpaceSizeName = "Disk space"

    def _reportDiskSpaceSize(self, reportRequest):
        return self.__reportFieldSizeinMB(reportRequest, ("DiskSpace", "Used disk space"))

    _reportInputDataSizeName = "Input data"

    def _reportInputDataSize(self, reportRequest):
        return self.__reportFieldSizeinMB(reportRequest, ("InputDataSize", "Input data"))

    _reportOutputDataSizeName = "Output data"

    def _reportOutputDataSize(self, reportRequest):
        return self.__reportFieldSizeinMB(reportRequest, ("OutputDataSize", "Output data"))

    def __reportFieldSizeinMB(self, reportRequest, fieldTuple):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", fieldTuple[0]],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, _maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "bytes"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotInputSandboxSize(self, reportRequest, plotInfo, filename):
        return self.__plotFieldSizeinMB(reportRequest, plotInfo, filename, ("InputSandBoxSize", "Input sand box size"))

    def _plotOutputSandboxSize(self, reportRequest, plotInfo, filename):
        return self.__plotFieldSizeinMB(
            reportRequest, plotInfo, filename, ("OutputSandBoxSize", "Output sand box size")
        )

    def _plotDiskSpaceSize(self, reportRequest, plotInfo, filename):
        return self.__plotFieldSizeinMB(reportRequest, plotInfo, filename, ("DiskSpace", "Used disk space"))

    def _plotInputDataSize(self, reportRequest, plotInfo, filename):
        return self.__plotFieldSizeinMB(reportRequest, plotInfo, filename, ("InputDataSize", "Input data"))

    def _plotOutputDataSize(self, reportRequest, plotInfo, filename):
        return self.__plotFieldSizeinMB(reportRequest, plotInfo, filename, ("OutputDataSize", "Output data"))

    def __plotFieldSizeinMB(self, reportRequest, plotInfo, filename, fieldTuple):
        metadata = {
            "title": "%s by %s" % (fieldTuple[1], reportRequest["grouping"]),
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateStackedLinePlot(filename, plotInfo["graphDataDict"], metadata)

    ##
    #  Cumulative data sizes
    ##

    _reportCumulativeInputSandboxSizeName = "Cumulative Input sandbox"

    def _reportCumulativeInputSandboxSize(self, reportRequest):
        return self.__reportCumulativeFieldSizeinMB(reportRequest, ("InputSandBoxSize", "Input sand box size"))

    _reportCumulativeOutputSandboxSizeName = "Cumulative Ouput sandbox"

    def _reportCumulativeOutputSandboxSize(self, reportRequest):
        return self.__reportCumulativeFieldSizeinMB(reportRequest, ("OutputSandBoxSize", "Output sand box size"))

    _reportCumulativeDiskSpaceSizeName = "Cumulative Disk space"

    def _reportCumulativeDiskSpaceSize(self, reportRequest):
        return self.__reportCumulativeFieldSizeinMB(reportRequest, ("DiskSpace", "Used disk space"))

    _reportCumulativeInputDataSizeName = "Cumulative Input data"

    def _reportCumulativeInputDataSize(self, reportRequest):
        return self.__reportCumulativeFieldSizeinMB(reportRequest, ("InputDataSize", "Input data"))

    _reportCumulativeOutputDataSizeName = "Cumulative Output data"

    def _reportCumulativeOutputDataSize(self, reportRequest):
        return self.__reportCumulativeFieldSizeinMB(reportRequest, ("OutputDataSize", "Output data"))

    def __reportCumulativeFieldSizeinMB(self, reportRequest, fieldTuple):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", fieldTuple[0]],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        dataDict = self._accumulate(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "bytes"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotCumulativeInputSandboxSize(self, reportRequest, plotInfo, filename):
        return self.__plotCumulativeFieldSizeinMB(
            reportRequest, plotInfo, filename, ("InputSandBoxSize", "Input sand box size")
        )

    def _plotCumulativeOutputSandboxSize(self, reportRequest, plotInfo, filename):
        return self.__plotCumulativeFieldSizeinMB(
            reportRequest, plotInfo, filename, ("OutputSandBoxSize", "Output sand box size")
        )

    def _plotCumulativeDiskSpaceSize(self, reportRequest, plotInfo, filename):
        return self.__plotCumulativeFieldSizeinMB(reportRequest, plotInfo, filename, ("DiskSpace", "Used disk space"))

    def _plotCumulativeInputDataSize(self, reportRequest, plotInfo, filename):
        return self.__plotCumulativeFieldSizeinMB(reportRequest, plotInfo, filename, ("InputDataSize", "Input data"))

    def _plotCumulativeOutputDataSize(self, reportRequest, plotInfo, filename):
        return self.__plotCumulativeFieldSizeinMB(reportRequest, plotInfo, filename, ("OutputDataSize", "Output data"))

    def __plotCumulativeFieldSizeinMB(self, reportRequest, plotInfo, filename, fieldTuple):
        metadata = {
            "title": "Cumulative %s by %s" % (fieldTuple[1], reportRequest["grouping"]),
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateCumulativePlot(filename, plotInfo["graphDataDict"], metadata)

    ##
    #  Input/Ouput data files
    ##

    _reportInputDataFilesName = "Input data files"

    def _reportInputDataFiles(self, reportRequest):
        return self.__reportDataFiles(reportRequest, ("InputDataFiles", "Input files"))

    _reportOuputDataFilesName = "Output data files"

    def _reportOuputDataFiles(self, reportRequest):
        return self.__reportDataFiles(reportRequest, ("OutputDataFiles", "Output files"))

    def __reportDataFiles(self, reportRequest, fieldTuple):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", fieldTuple[0]],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, _maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "files"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotInputDataFiles(self, reportRequest, plotInfo, filename):
        return self.__plotDataFiles(reportRequest, plotInfo, filename, ("InputDataFiles", "Input files"))

    def _plotOuputDataFiles(self, reportRequest, plotInfo, filename):
        return self.__plotDataFiles(reportRequest, plotInfo, filename, ("OutputDataFiles", "Output files"))

    def __plotDataFiles(self, reportRequest, plotInfo, filename, fieldTuple):
        metadata = {
            "title": "%s by %s" % (fieldTuple[1], reportRequest["grouping"]),
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateStackedLinePlot(filename, plotInfo["graphDataDict"], metadata)

    _reportHistogramCPUUsedName = "Histogram CPU time"

    def _reportHistogramCPUUsed(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s",
            reportRequest["groupingFields"][1] + ["CPUTime"],
        )

        retVal = self._getBucketData(
            reportRequest["startTime"], reportRequest["endTime"], selectFields, reportRequest["condDict"]
        )
        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]
        return S_OK({"data": dataDict})

    def _plotHistogramCPUUsed(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "CPU usage by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
        }
        return self._generateHistogram(filename, [plotInfo["data"]], metadata)



########################################################################
# File: DIRACSingleton.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2011/11/24 13:40:38
########################################################################

"""
:mod: DIRACSingleton

.. module: DIRACSingleton

:synopsis: metaclass singleton implementation

.. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

metaclass singleton implementation

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

##
# @file DIRACSingleton.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2011/11/24 13:40:49
# @brief Definition of DIRACSingleton class.


class DIRACSingleton(type):
    """
    Simple singleton pattern using metaclass

    If you want make your class a singleton, just set its  __metaclass__ to
    DIRACSingleton, i.e.::

      import six
      from DIRAC.Core.Utilities.DIRACSingleton import DIRACSingleton
      @six.add_metaclass(DIRACSingleton)
      class CheesShop(object):
        ...
    """

    def __init__(cls, name, bases, dic):
        """c'tor

        :param cls: class def
        :param name: class name (becomes __name__ attr)
        :param bases: tuple of parent class defs (becomes __bases__ attr)
        :param dic: definition dict for class body (becomes __dict__ attr)
        """
        super(DIRACSingleton, cls).__init__(name, bases, dic)
        cls.instance = None

    def __call__(cls, *args, **kwargs):
        """
        Get the only one instance of cls

        :param cls: class def
        :param list args: anon args list
        :param dict kwargs: named args dict
        """
        if cls.instance is None:
            cls.instance = super(DIRACSingleton, cls).__call__(*args, **kwargs)
        return cls.instance



"""Module collecting functions dealing with the GLUE2 information schema

:author: A.Sailer

Known problems:

 * ARC CEs do not seem to publish wall or CPU time per queue anywhere
 * There is no consistency between which memory information is provided where,
   execution environment vs. information for a share
 * Some execution environment IDs are used more than once

Print outs with "SCHEMA PROBLEM" point -- in my opinion -- to errors in the
published information, like a foreign key pointing to non-existent entry.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from pprint import pformat

from DIRAC import gLogger, gConfig
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCESiteMapping, getGOCSiteName
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.Grid import ldapsearchBDII

__RCSID__ = "$Id$"


sLog = gLogger.getSubLogger(__name__)


def getGlue2CEInfo(vo, host=None):
    """call ldap for GLUE2 and get information

    :param str vo: Virtual Organisation
    :param str host: host to query for information
    :returns: result structure with result['Value'][siteID]['CEs'][ceID]['Queues'][queueName]. For
                 each siteID, ceID, queueName all the GLUE2 parameters are retrieved
    """

    # get all Policies allowing given VO
    filt = "(&(objectClass=GLUE2Policy)(|(GLUE2PolicyRule=VO:%s)(GLUE2PolicyRule=vo:%s)))" % (vo, vo)
    polRes = ldapsearchBDII(filt=filt, attr=None, host=host, base="o=glue", selectionString="GLUE2")

    if not polRes["OK"]:
        return S_ERROR("Failed to get policies for this VO")
    polRes = polRes["Value"]

    sLog.notice("Found %s policies for this VO %s" % (len(polRes), vo))
    # get all shares for this policy
    # create an or'ed list of all the shares and then call the search
    listOfSitesWithPolicies = set()
    shareFilter = ""
    for policyValues in polRes:
        # skip entries without GLUE2DomainID in the DN because we cannot associate them to a site
        if "GLUE2DomainID" not in policyValues["attr"]["dn"]:
            continue
        shareID = policyValues["attr"].get("GLUE2MappingPolicyShareForeignKey", None)
        policyID = policyValues["attr"]["GLUE2PolicyID"]
        siteName = policyValues["attr"]["dn"].split("GLUE2DomainID=")[1].split(",", 1)[0]
        listOfSitesWithPolicies.add(siteName)
        if shareID is None:  # policy not pointing to ComputingInformation
            sLog.debug("Policy %s does not point to computing information" % (policyID,))
            continue
        sLog.verbose("%s policy %s pointing to %s " % (siteName, policyID, shareID))
        sLog.debug("Policy values:\n%s" % pformat(policyValues))
        shareFilter += "(GLUE2ShareID=%s)" % shareID

    filt = "(&(objectClass=GLUE2Share)(|%s))" % shareFilter
    shareRes = ldapsearchBDII(filt=filt, attr=None, host=host, base="o=glue", selectionString="GLUE2")
    if not shareRes["OK"]:
        sLog.error("Could not get share information", shareRes["Message"])
        return shareRes
    shareInfoLists = {}
    for shareInfo in shareRes["Value"]:
        if "GLUE2DomainID" not in shareInfo["attr"]["dn"]:
            continue
        if "GLUE2ComputingShare" not in shareInfo["objectClass"]:
            sLog.debug("Share %r is not a ComputingShare: \n%s" % (shareID, pformat(shareInfo)))
            continue
        sLog.debug("Found computing share:\n%s" % pformat(shareInfo))
        siteName = shareInfo["attr"]["dn"].split("GLUE2DomainID=")[1].split(",", 1)[0]
        shareInfoLists.setdefault(siteName, []).append(shareInfo["attr"])

    siteInfo = __getGlue2ShareInfo(host, shareInfoLists)
    if not siteInfo["OK"]:
        sLog.error("Could not get CE info for", "%s: %s" % (shareID, siteInfo["Message"]))
        return siteInfo
    siteDict = siteInfo["Value"]
    sLog.debug("Found Sites:\n%s" % pformat(siteDict))
    sitesWithoutShares = set(siteDict) - listOfSitesWithPolicies
    if sitesWithoutShares:
        sLog.error("Found some sites without any shares", pformat(sitesWithoutShares))
    else:
        sLog.notice("Found information for all known sites")

    # remap siteDict to assign CEs to known sites,
    # in case their names differ from the "gocdb name" in the CS.
    newSiteDict = {}
    ceSiteMapping = getCESiteMapping().get("Value", {})
    # pylint thinks siteDict is a tuple, so we cast
    for siteName, infoDict in dict(siteDict).items():
        for ce, ceInfo in infoDict.get("CEs", {}).items():
            ceSiteName = ceSiteMapping.get(ce, siteName)
            gocSiteName = getGOCSiteName(ceSiteName).get("Value", siteName)
            newSiteDict.setdefault(gocSiteName, {}).setdefault("CEs", {})[ce] = ceInfo

    return S_OK(newSiteDict)


def __getGlue2ShareInfo(host, shareInfoLists):
    """get information from endpoints, which are the CE at a Site

    :param str host: BDII host to query
    :param dict shareInfoDict: dictionary of GLUE2 parameters belonging to the ComputingShare
    :returns: result structure S_OK/S_ERROR
    """
    executionEnvironments = []
    for _siteName, shareInfoDicts in shareInfoLists.items():
        for shareInfoDict in shareInfoDicts:
            executionEnvironment = shareInfoDict.get("GLUE2ComputingShareExecutionEnvironmentForeignKey", [])
            if not executionEnvironment:
                sLog.error("No entry for GLUE2ComputingShareExecutionEnvironmentForeignKey", pformat(shareInfoDict))
                continue
            if isinstance(executionEnvironment, six.string_types):
                executionEnvironment = [executionEnvironment]
            executionEnvironments.extend(executionEnvironment)
    resExeInfo = __getGlue2ExecutionEnvironmentInfo(host, executionEnvironments)
    if not resExeInfo["OK"]:
        sLog.error(
            "Cannot get execution environment info for:",
            str(executionEnvironments)[:100] + "  " + resExeInfo["Message"],
        )
        return resExeInfo
    exeInfos = resExeInfo["Value"]

    siteDict = {}
    for siteName, shareInfoDicts in shareInfoLists.items():
        siteDict[siteName] = {"CEs": {}}
        cesDict = siteDict[siteName]["CEs"]
        for shareInfoDict in shareInfoDicts:
            ceInfo = {}
            ceInfo["MaxWaitingJobs"] = shareInfoDict.get("GLUE2ComputingShareMaxWaitingJobs", "-1")  # This is not used
            ceInfo["Queues"] = {}
            queueInfo = {}
            queueInfo["GlueCEStateStatus"] = shareInfoDict["GLUE2ComputingShareServingState"]
            queueInfo["GlueCEPolicyMaxCPUTime"] = str(
                int(int(shareInfoDict.get("GLUE2ComputingShareMaxCPUTime", 86400)) / 60)
            )
            queueInfo["GlueCEPolicyMaxWallClockTime"] = str(
                int(int(shareInfoDict.get("GLUE2ComputingShareMaxWallTime", 86400)) / 60)
            )
            queueInfo["GlueCEInfoTotalCPUs"] = shareInfoDict.get("GLUE2ComputingShareMaxRunningJobs", "10000")
            queueInfo["GlueCECapability"] = ["CPUScalingReferenceSI00=2552"]

            try:
                maxNOPfromCS = gConfig.getValue(
                    "/Resources/Computing/CEDefaults/GLUE2ComputingShareMaxSlotsPerJob_limit", 8
                )
                maxNOPfromGLUE = int(shareInfoDict.get("GLUE2ComputingShareMaxSlotsPerJob", 1))
                numberOfProcs = min(maxNOPfromGLUE, maxNOPfromCS)
                queueInfo["NumberOfProcessors"] = numberOfProcs
                if numberOfProcs != maxNOPfromGLUE:
                    sLog.info(
                        "Limited NumberOfProcessors for", "%s from %s to %s" % (siteName, maxNOPfromGLUE, numberOfProcs)
                    )
            except ValueError:
                sLog.error(
                    "Bad content for GLUE2ComputingShareMaxSlotsPerJob:",
                    siteName + " " + shareInfoDict.get("GLUE2ComputingShareMaxSlotsPerJob"),
                )
                queueInfo["NumberOfProcessors"] = 1

            executionEnvironment = shareInfoDict.get("GLUE2ComputingShareExecutionEnvironmentForeignKey", [])
            if isinstance(executionEnvironment, six.string_types):
                executionEnvironment = [executionEnvironment]
            resExeInfo = __getGlue2ExecutionEnvironmentInfoForSite(siteName, executionEnvironment, exeInfos)
            if not resExeInfo["OK"]:
                continue

            exeInfo = resExeInfo.get("Value")
            if not exeInfo:
                sLog.error("Using dummy values. Did not find information for execution environment", siteName)
                exeInfo = {
                    "GlueHostMainMemoryRAMSize": "1999",  # intentionally identifiably dummy value
                    "GlueHostOperatingSystemVersion": "",
                    "GlueHostOperatingSystemName": "",
                    "GlueHostOperatingSystemRelease": "",
                    "GlueHostArchitecturePlatformType": "x86_64",
                    "GlueHostBenchmarkSI00": "2500",  # needed for the queue to be used by the sitedirector
                    "MANAGER": "manager:unknownBatchSystem",  # need some value for ARC
                }
            else:
                sLog.info("Found information for execution environment for", siteName)

            # sometimes the time is still in hours
            maxCPUTime = int(queueInfo["GlueCEPolicyMaxCPUTime"])
            if maxCPUTime in [12, 24, 36, 48, 168]:
                queueInfo["GlueCEPolicyMaxCPUTime"] = str(maxCPUTime * 60)
                queueInfo["GlueCEPolicyMaxWallClockTime"] = str(int(queueInfo["GlueCEPolicyMaxWallClockTime"]) * 60)

            ceInfo.update(exeInfo)
            shareEndpoints = shareInfoDict.get("GLUE2ShareEndpointForeignKey", [])
            if isinstance(shareEndpoints, six.string_types):
                shareEndpoints = [shareEndpoints]
            for endpoint in shareEndpoints:
                ceType = endpoint.rsplit(".", 1)[1]
                # get queue Name, in CREAM this is behind GLUE2entityOtherInfo...
                if ceType == "CREAM":
                    for otherInfo in shareInfoDict["GLUE2EntityOtherInfo"]:
                        if otherInfo.startswith("CREAMCEId"):
                            queueName = otherInfo.split("/", 1)[1]
                            # creamCEs are EOL soon, ignore any info they have
                            if queueInfo.pop("NumberOfProcessors", 1) != 1:
                                sLog.verbose("Ignoring MaxSlotsPerJob option for CreamCE", endpoint)

                # HTCondorCE, htcondorce
                elif ceType.lower().endswith("htcondorce"):
                    ceType = "HTCondorCE"
                    queueName = "condor"

                else:
                    sLog.error("Unknown CE Type, please check the available information", ceType)
                    continue

                queueInfo["GlueCEImplementationName"] = ceType
                ceName = endpoint.split("_", 1)[0]
                cesDict.setdefault(ceName, {})
                existingQueues = dict(cesDict[ceName].get("Queues", {}))
                existingQueues[queueName] = queueInfo
                ceInfo["Queues"] = existingQueues
                cesDict[ceName].update(ceInfo)

            # ARC CEs do not have endpoints, we have to try something else to get the information about the queue etc.
            try:
                if not shareEndpoints and shareInfoDict["GLUE2ShareID"].startswith("urn:ogf"):
                    exeInfo = dict(exeInfo)  # silence pylint about tuples
                    queueInfo["GlueCEImplementationName"] = "ARC"
                    managerName = exeInfo.pop("MANAGER", "").split(" ", 1)[0].rsplit(":", 1)[1]
                    managerName = managerName.capitalize() if managerName == "condor" else managerName
                    queueName = "nordugrid-%s-%s" % (managerName, shareInfoDict["GLUE2ComputingShareMappingQueue"])
                    ceName = shareInfoDict["GLUE2ShareID"].split("ComputingShare:")[1].split(":")[0]
                    cesDict.setdefault(ceName, {})
                    existingQueues = dict(cesDict[ceName].get("Queues", {}))
                    existingQueues[queueName] = queueInfo
                    ceInfo["Queues"] = existingQueues
                    cesDict[ceName].update(ceInfo)
            except Exception:
                sLog.error("Exception in ARC part for site:", siteName)

    return S_OK(siteDict)


def __getGlue2ExecutionEnvironmentInfo(host, executionEnvironments):
    """Find all the executionEnvironments.

    :param str host: BDII host to query
    :param list executionEnvironments: list of the execution environments to get some information from
    :returns: result of the ldapsearch for all executionEnvironments, Glue2 schema
    """
    listOfValues = []
    # break up to avoid argument list too long, it started failing at about 1900 entries
    for exeEnvs in breakListIntoChunks(executionEnvironments, 1000):
        exeFilter = ""
        for execEnv in exeEnvs:
            exeFilter += "(GLUE2ResourceID=%s)" % execEnv
        filt = "(&(objectClass=GLUE2ExecutionEnvironment)(|%s))" % exeFilter
        response = ldapsearchBDII(filt=filt, attr=None, host=host, base="o=glue", selectionString="GLUE2")
        if not response["OK"]:
            return response
        if not response["Value"]:
            sLog.error("No information found for %s" % executionEnvironments)
            continue
        listOfValues += response["Value"]
    if not listOfValues:
        return S_ERROR("No information found for executionEnvironments")
    return S_OK(listOfValues)


def __getGlue2ExecutionEnvironmentInfoForSite(sitename, foreignKeys, exeInfos):
    """Get the information about the execution environment for a specific site or ce or something.

    :param str sitename: Name of the site we are looking at
    :param list foreignKeys: list of ExecutionEnvironmentForeignkeys linked by the site
    :param list exeInfos: bdii list of dictionaries containing all the ExecutionEnvironment information for all sites
    :return: Dictionary with the information as required by the Bdii2CSagent for this site
    """
    # filter those that we want
    exeInfos = [exeInfo for exeInfo in exeInfos if exeInfo["attr"]["GLUE2ResourceID"] in foreignKeys]
    # take the CE with the lowest MainMemory
    exeInfo = sorted(exeInfos, key=lambda k: int(k["attr"]["GLUE2ExecutionEnvironmentMainMemorySize"]))
    if not exeInfo:
        sLog.error(
            "SCHEMA PROBLEM: Did not find execution info for site", sitename + " and keys: " + " ".join(foreignKeys)
        )
        return S_OK()
    sLog.debug("Found ExecutionEnvironments", pformat(exeInfo[0]))
    exeInfo = exeInfo[0]["attr"]  # pylint: disable=unsubscriptable-object
    maxRam = exeInfo.get("GLUE2ExecutionEnvironmentMainMemorySize", "")
    architecture = exeInfo.get("GLUE2ExecutionEnvironmentPlatform", "")
    architecture = "x86_64" if architecture == "amd64" else architecture
    architecture = "x86_64" if architecture == "UNDEFINEDVALUE" else architecture
    architecture = "x86_64" if "Intel(R) Xeon(R)" in architecture else architecture
    osFamily = exeInfo.get("GLUE2ExecutionEnvironmentOSFamily", "")  # e.g. linux
    osName = exeInfo.get("GLUE2ExecutionEnvironmentOSName", "")
    osVersion = exeInfo.get("GLUE2ExecutionEnvironmentOSVersion", "")
    manager = exeInfo.get("GLUE2ExecutionEnvironmentComputingManagerForeignKey", "manager:unknownBatchSystem")
    # translate to Glue1 like keys, because that is used later on
    infoDict = {
        "GlueHostMainMemoryRAMSize": maxRam,
        "GlueHostOperatingSystemVersion": osName,
        "GlueHostOperatingSystemName": osFamily,
        "GlueHostOperatingSystemRelease": osVersion,
        "GlueHostArchitecturePlatformType": architecture.lower(),
        "GlueHostBenchmarkSI00": "2500",  # needed for the queue to be used by the sitedirector
        "MANAGER": manager,  # to create the ARC QueueName mostly
    }

    return S_OK(infoDict)
